"""
Module containing geometry block set operations
"""
import geopandas as gpd
from shapely.geometry import box

from .base import GeometryBlock, BaseSingle

__all__ = ["Difference", "Intersection"]


class Difference(BaseSingle):
    """Block that calculates the difference of two GeometryBlocks.

    The resulting GeometryBlock will have all geometries in ``source``, and if
    there are geometries with the same ID in ``other``, the geometries will be
    adapted using the Difference operation.

    :type source: GeometryBlock
    :type other: GeometryBlock
    """

    def __init__(self, source, other):
        if not isinstance(other, GeometryBlock):
            raise TypeError("'{}' object is not allowed".format(type(other)))
        super().__init__(source, other)  # type of source is checked in super()

    @property
    def other(self):
        return self.args[1]

    def get_sources_and_requests(self, **request):
        if request["mode"] == "extent":
            # the resulting extent might be smaller than the source's extent,
            # but we would have to execute the Difference in order to know.
            return [(self.source, request)]

        # we need to get the extent of the geometries in source, in order to
        # fetch all needed geometries in other
        extent_request = request.copy()
        extent_request["mode"] = "extent"
        extent = self.source.get_data(**extent_request)["extent"]

        if extent is None:  # shortcut when there are no geometries present
            projection = request["projection"]
            return [({"empty": True, "projection": projection}, None)]

        other_request = request.copy()
        other_request["geometry"] = box(*extent)
        return [(self.source, request), (self.other, other_request)]

    @staticmethod
    def process(source_data, other_data=None):
        if other_data is None:
            if source_data.get("empty"):  # catches the no-geometry shortcut
                return {
                    "features": gpd.GeoDataFrame([]),
                    "projection": source_data["projection"],
                }
            else:
                return source_data

        a = source_data["features"]
        b = other_data["features"]
        if len(a) == 0 or len(b) == 0:
            return source_data  # do nothing

        a_series = a["geometry"]
        b_series = b["geometry"].reindex(a_series.index)
        result_series = a_series.difference(b_series)

        # re-insert geometries that were missing in b (we want A - None = A)
        missing_in_b = b_series.isna()
        result_series[missing_in_b] = a_series[missing_in_b]
        result = a.set_geometry(result_series)
        return {"features": result, "projection": source_data["projection"]}


class Intersection(BaseSingle):
    """Block that intersects geometries with the requested geometry.

    :param source: the source of geometry data
    :type source: GeometryBlock
    """

    # TODO There are three modes, of which only one is currently implemented:
    #
    # 1. Intersection of source geometries with the requested geometry.
    # 2. Intersection of source geometries with a single, constant geometry,
    # 3. Intersection of source geometries with geometries from another source.
    #
    # Only case 1 is implemented. Supply ``other=None`` (default).
    def __init__(self, source, other=None):
        if isinstance(other, GeometryBlock):
            raise NotImplementedError(
                "Cannot compare geometries with another geometry datasource"
            )
        elif other is not None:
            raise NotImplementedError(
                "Cannot compare geometries with a constant geometry"
            )
        super().__init__(source, other)

    @property
    def other(self):
        return self.args[1]

    def get_sources_and_requests(self, **request):
        return [(self.source, request), (request["geometry"], None)]

    @staticmethod
    def process(data, geometry):
        # the features will always be in the projection of geometry so we can
        # safely not think about projections here.
        # This cover 'intersects' and 'centroid' mode.
        if "features" in data:
            features = data["features"]
            features = features.set_geometry(features.geometry.intersection(geometry))
            return {"features": features, "projection": data["projection"]}
        elif "extent" in data:
            # intersect the bboxes
            bbox1 = data["extent"]
            bbox2 = geometry.bounds
            bbox = (
                max(bbox1[0], bbox2[0]),
                max(bbox1[1], bbox2[1]),
                min(bbox1[2], bbox2[2]),
                min(bbox1[3], bbox2[3]),
            )
            return {"extent": bbox, "projection": data["projection"]}



"""
Module containing raster blocks that aggregate rasters.
"""
from math import ceil, floor, log, sqrt
from collections import defaultdict
from functools import partial
import warnings

from scipy import ndimage
import numpy as np
import geopandas as gpd

from dask import config
from dask_geomodeling import measurements
from dask_geomodeling import utils
from dask_geomodeling.raster import RasterBlock

from .base import GeometryBlock

__all__ = ["AggregateRaster", "AggregateRasterAboveThreshold"]


class Bucket:
    """
    Track objects in an imaginary grid that may span up to 4 cells.
    """

    def __init__(self):
        self.cells = set()
        self.indices = []

    def __contains__(self, cells):
        """
        Return wether any of the cells defined by indices is already occupied.
        """
        return bool(self.cells & cells)

    def add(self, index, cells):
        """
        Update the set of occupied cells with cells and append index to the
        list of indices.

        Note that this does not fail if cells intersects with already occupied
        cells.
        """
        self.indices.append(index)
        self.cells.update(cells)


def calculate_level_and_cells(bbox):
    """
    Return a tuple (level, cells).

    :param bboxes: list of (xmin, ymin, xmax, ymax) tuples

    The returned cells is a set of indices which represent occupied cells (at
    most 4) in an imaginary sparse grid that has a cellsize defined by the
    integer level.  Level 0 corresponds to the unit cell. Each doubling of the
    cellsize level increase corresponds to a doubling of the cellsize of the
    previous level.
    """
    x1, y1, x2, y2 = bbox
    level = -ceil(log(max(x2 - x1, y2 - y1), 2))

    width = 0.5 ** level
    height = 0.5 ** level

    j1 = floor(x1 / width)
    j2 = floor(x2 / width)
    i1 = floor(y1 / height)
    i2 = floor(y2 / height)

    return level, {(i1, j1), (i1, j2), (i2, j1), (i2, j2)}


def bucketize(bboxes):
    """
    Return list of lists with indices into bboxes.

    :param bboxes: list of (xmin, ymin, xmax, ymax) tuples

    Each sublist in the returned list points to a subset of disjoint bboxes.
    Instead of aiming for the smallest amount of subsets possible, this
    approach focuses on speed by avoiding costly intersection operations on all
    bboxes in a bucket.
    """
    bucket_dict = defaultdict(list)

    for index, bbox in enumerate(bboxes):
        level, cells = calculate_level_and_cells(bbox)

        # select bucket by feature size
        bucket_list = bucket_dict[level]

        for bucket in bucket_list:
            # see if it fits in this bucket
            if cells in bucket:
                continue
            # a suitable bucket has been found, break out of the for loop
            break
        else:
            # no break, no suitable bucket found, assign and append a new one
            bucket = Bucket()
            bucket_list.append(bucket)

        # add the item to the bucket
        bucket.add(index=index, cells=cells)

    return [
        bucket.indices for bucket_list in bucket_dict.values() for bucket in bucket_list
    ]


class AggregateRaster(GeometryBlock):
    """
    Compute zonal statistics and add them to the geometry properties

    :param source: the source of geometry data
    :param raster: the source of raster data
    :param statistic: the type of statistic to perform. can be
      ``'sum', 'count', 'min', 'max', 'mean', 'median', 'p<percentile>'``.
    :param projection: the projection to perform the aggregation in
    :param pixel_size: the pixel size to perform aggregation in
    :param max_pixels: the maximum number of pixels to use for aggregation.
       defaults to the geomodeling.raster-limit setting.
    :param column_name: the name of the column to output the results
    :param auto_pixel_size: determines whether the pixel_size is
      adjusted when a raster is too large. Default False.
    :returns: GeometryBlock with aggregation results in ``column_name``

    :type source: GeometryBlock
    :type raster: RasterBlock
    :type statistic: string
    :type projection: string or None
    :type pixel_size: float or None
    :type max_pixels: int or None
    :type column_name: string
    :type auto_pixel_size: boolean

    The currently implemented statistics are sum, count, min, max, mean,
    median, and percentile. If projection or max_resolution are not
    given, these are taken from the provided RasterBlock.

    The count statistic calculates the number of active cells in the raster. A
    percentile statistic can be selected using text value starting with 'p'
    followed by something that can be parsed as a float value, for example
    ``'p33.3'``.

    Only geometries that intersect the requested bbox are aggregated.
    Aggregation is done in a specified projection and with a specified pixel
    size.

    Should the combination of the requested pixel_size and the extent of the
    source geometry cause the requested raster size to exceed max_pixels, the
    pixel_size is adjusted automatically if ``auto_pixel_size = True``, else
    a RuntimeError is raised.

    The global raster-limit setting can be adapted as follows:
      >>> from dask import config
      >>> config.set({"geomodeling.raster-limit": 10 ** 9})
    """

    # extensive (opposite: intensive) means: additive, proportional to size
    STATISTICS = {
        "sum": {"func": ndimage.sum, "extensive": True},
        "count": {"func": ndimage.sum, "extensive": True},
        "min": {"func": ndimage.minimum, "extensive": False},
        "max": {"func": ndimage.maximum, "extensive": False},
        "mean": {"func": ndimage.mean, "extensive": False},
        "median": {"func": ndimage.median, "extensive": False},
        "percentile": {"func": measurements.percentile, "extensive": False},
    }

    def __init__(
        self,
        source,
        raster,
        statistic="sum",
        projection=None,
        pixel_size=None,
        max_pixels=None,
        column_name="agg",
        auto_pixel_size=False,
        *args
    ):
        if not isinstance(source, GeometryBlock):
            raise TypeError("'{}' object is not allowed".format(type(source)))
        if not isinstance(raster, RasterBlock):
            raise TypeError("'{}' object is not allowed".format(type(raster)))
        if not isinstance(statistic, str):
            raise TypeError("'{}' object is not allowed".format(type(statistic)))
        statistic = statistic.lower()
        percentile = utils.parse_percentile_statistic(statistic)
        if percentile:
            statistic = "p{0}".format(percentile)
        elif statistic not in self.STATISTICS or statistic == "percentile":
            raise ValueError("Unknown statistic '{}'".format(statistic))

        if projection is None:
            projection = raster.projection
        if not isinstance(projection, str):
            raise TypeError("'{}' object is not allowed".format(type(projection)))
        if pixel_size is None:
            # get the pixel_size from the raster geo_transform
            geo_transform = raster.geo_transform
            if geo_transform is None:
                raise ValueError(
                    "Cannot get the pixel_size from the source "
                    "raster. Please provide a pixel_size."
                )
            pixel_size = min(abs(float(geo_transform[1])), abs(float(geo_transform[5])))
        else:
            pixel_size = abs(float(pixel_size))
        if pixel_size == 0.0:
            raise ValueError("Pixel size cannot be 0")
        if max_pixels is not None:
            max_pixels = int(max_pixels)
        if not isinstance(auto_pixel_size, bool):
            raise TypeError("'{}' object is not allowed".format(type(auto_pixel_size)))

        super(AggregateRaster, self).__init__(
            source,
            raster,
            statistic,
            projection,
            pixel_size,
            max_pixels,
            column_name,
            auto_pixel_size,
            *args
        )

    @property
    def source(self):
        return self.args[0]

    @property
    def raster(self):
        return self.args[1]

    @property
    def statistic(self):
        return self.args[2]

    @property
    def projection(self):
        return self.args[3]

    @property
    def pixel_size(self):
        return self.args[4]

    @property
    def max_pixels(self):
        return self.args[5]

    @property
    def column_name(self):
        return self.args[6]

    @property
    def auto_pixel_size(self):
        return self.args[7]

    @property
    def columns(self):
        return self.source.columns | {self.column_name}

    def get_sources_and_requests(self, **request):
        if request.get("mode") == "extent":
            return [(self.source, request), (None, None), ({"mode": "extent"}, None)]

        req_srs = request["projection"]
        agg_srs = self.projection

        # acquire the extent of the geometry data
        extent_request = {**request, "mode": "extent"}
        extent = self.source.get_data(**extent_request)["extent"]

        if extent is None:
            # make sources_and_request so that we get an empty result
            return [
                (None, None),
                (None, None),
                ({"empty": True, "projection": req_srs}, None),
            ]

        # transform the extent into the projection in which we aggregate
        x1, y1, x2, y2 = utils.transform_extent(extent, req_srs, agg_srs)

        # estimate the amount of required pixels
        required_pixels = int(((x2 - x1) * (y2 - y1)) / (self.pixel_size ** 2))

        # in case this request is too large, we adapt pixel size
        max_pixels = self.max_pixels
        if max_pixels is None:
            max_pixels = config.get("geomodeling.raster-limit")
        pixel_size = self.pixel_size

        if required_pixels > max_pixels and self.auto_pixel_size:
            # adapt with integer multiples of pixel_size
            pixel_size *= ceil(sqrt(required_pixels / max_pixels))
        elif required_pixels > max_pixels:
            raise RuntimeError(
                "The required raster size for the aggregation exceeded "
                "the maximum ({} > {})".format(required_pixels, max_pixels)
            )

        # snap the extent to (0, 0) to prevent subpixel shifts
        x1 = floor(x1 / pixel_size) * pixel_size
        y1 = floor(y1 / pixel_size) * pixel_size
        x2 = ceil(x2 / pixel_size) * pixel_size
        y2 = ceil(y2 / pixel_size) * pixel_size

        # compute the width and height
        width = max(int((x2 - x1) / pixel_size), 1)
        height = max(int((y2 - y1) / pixel_size), 1)

        raster_request = {
            "mode": "vals",
            "projection": agg_srs,
            "start": request.get("start"),
            "stop": request.get("stop"),
            "aggregation": None,  # TODO
            "bbox": (x1, y1, x2, y2),
            "width": width,
            "height": height,
        }

        process_kwargs = {
            "mode": request.get("mode", "intersects"),
            "pixel_size": self.pixel_size,
            "agg_srs": agg_srs,
            "req_srs": req_srs,
            "actual_pixel_size": pixel_size,
            "statistic": self.statistic,
            "result_column": self.column_name,
            "agg_bbox": (x1, y1, x2, y2),
        }

        return [
            (self.source, request),
            (self.raster, raster_request),
            (process_kwargs, None),
        ]

    @staticmethod
    def process(geom_data, raster_data, process_kwargs):
        if process_kwargs.get("empty"):
            return {
                "features": gpd.GeoDataFrame([]),
                "projection": process_kwargs["projection"],
            }
        elif process_kwargs["mode"] == "extent":
            return geom_data

        features = geom_data["features"]
        if len(features) == 0:
            return geom_data

        result = features.copy()

        # transform the features into the aggregation projection
        req_srs = process_kwargs["req_srs"]
        agg_srs = process_kwargs["agg_srs"]

        agg_geometries = utils.geoseries_transform(
            features["geometry"], req_srs, agg_srs,
        )

        statistic = process_kwargs["statistic"]
        percentile = utils.parse_percentile_statistic(statistic)
        if percentile:
            statistic = "percentile"
            agg_func = partial(
                AggregateRaster.STATISTICS[statistic]["func"], qval=percentile
            )
        else:
            agg_func = AggregateRaster.STATISTICS[statistic]["func"]

        extensive = AggregateRaster.STATISTICS[statistic]["extensive"]
        result_column = process_kwargs["result_column"]

        # this is only there for the AggregateRasterAboveThreshold
        threshold_name = process_kwargs.get("threshold_name")

        # investigate the raster data
        if raster_data is None:
            values = no_data_value = None
        else:
            values = raster_data["values"]
            no_data_value = raster_data["no_data_value"]
        if values is None or np.all(values == no_data_value):  # skip the rest
            result[result_column] = 0 if extensive else np.nan
            return {"features": result, "projection": req_srs}
        depth, height, width = values.shape

        pixel_size = process_kwargs["pixel_size"]
        actual_pixel_size = process_kwargs["actual_pixel_size"]

        # process in groups of disjoint subsets of the features
        agg = np.full((depth, len(features)), np.nan, dtype="f4")
        for select in bucketize(features.bounds.values):
            agg_geometries_bucket = agg_geometries.iloc[select]
            index = features.index[select]

            rasterize_result = utils.rasterize_geoseries(
                agg_geometries_bucket,
                process_kwargs["agg_bbox"],
                agg_srs,
                height,
                width,
                values=index,
            )
            labels = rasterize_result["values"][0]

            # if there is a threshold, generate a raster with thresholds
            if threshold_name:
                thresholds = features.loc[
                    labels.ravel(), threshold_name
                ].values.reshape(labels.shape)
            else:
                thresholds = None

            for frame_no, frame in enumerate(values):
                # limit statistics to active pixels
                active = frame != no_data_value
                # if there is a threshold, mask the frame
                if threshold_name:
                    valid = ~np.isnan(thresholds)  # to suppress warnings
                    active[~valid] = False  # no threshold -> no aggregation
                    active[valid] &= frame[valid] >= thresholds[valid]

                # if there is no single active value: do not aggregate
                if not active.any():
                    continue

                with warnings.catch_warnings():
                    # we may get divide by 0 if any geometry does not contain
                    # any 'active' values
                    warnings.simplefilter("ignore")
                    agg[frame_no][select] = agg_func(
                        1 if statistic == "count" else frame[active],
                        labels=labels[active],
                        index=index,
                    )

        if extensive:  # sum and count
            agg[~np.isfinite(agg)] = 0
            # extensive aggregations have to be scaled
            if actual_pixel_size != pixel_size:
                agg *= (actual_pixel_size / pixel_size) ** 2
        else:
            agg[~np.isfinite(agg)] = np.nan  # replaces inf by nan

        if depth == 1:
            result[result_column] = agg[0]
        else:
            # store an array in a dataframe cell: set each cell with [np.array]
            result[result_column] = [[x] for x in agg.T]

        return {"features": result, "projection": req_srs}


class AggregateRasterAboveThreshold(AggregateRaster):
    """
    Aggregate raster values ignoring values below some threshold. The
    thresholds are supplied per geometry.

    :param source: the source of geometry data
    :param raster: the source of raster data
    :param statistic: the type of statistic to perform. can be
      ``'sum', 'count', 'min', 'max', 'mean', 'median', 'p<percentile>'``.
    :param projection: the projection to perform the aggregation in
    :param pixel_size: the pixel size to perform aggregation in
    :param max_pixels: the maximum number of pixels to use for aggregation
    :param column_name: the name of the column to output the results
    :param auto_pixel_size: determines whether the pixel_size is
      adjusted when a raster is too large. Default False.
    :param threshold_name: the name of the column with the thresholds
    :returns: GeometryBlock with aggregation results in ``column_name``

    :type source: GeometryBlock
    :type raster: RasterBlock
    :type statistic: string
    :type projection: string
    :type pixel_size: float
    :type max_pixels: int
    :type column_name: string
    :type auto_pixel_size: boolean
    :type threshold_name: string

    See also:
      :class:`dask_geomodeling.geometry.aggregate.AggregateRaster`
    """

    def __init__(
        self,
        source,
        raster,
        statistic="sum",
        projection=None,
        pixel_size=None,
        max_pixels=None,
        column_name="agg",
        auto_pixel_size=False,
        threshold_name=None,
    ):
        if not isinstance(threshold_name, str):
            raise TypeError("'{}' object is not allowed".format(type(threshold_name)))
        if threshold_name not in source.columns:
            raise KeyError("Column '{}' is not available".format(threshold_name))
        super().__init__(
            source,
            raster,
            statistic,
            projection,
            pixel_size,
            max_pixels,
            column_name,
            auto_pixel_size,
            threshold_name,
        )

    @property
    def threshold_name(self):
        return self.args[8]

    def get_sources_and_requests(self, **request):
        src_and_req = super().get_sources_and_requests(**request)
        process_kwargs = src_and_req[2][0]
        process_kwargs["threshold_name"] = self.threshold_name
        return src_and_req



"""
Module containing text column operations that act on geometry blocks
"""
import pandas as pd
import re

from .base import GeometryBlock, BaseSingle

__all__ = ["ParseTextColumn"]

# https://catonmat.net/my-favorite-regex
# key matches any ASCII char except for '='
# value matches any ASCII char
REGEX_KEYVALUE = re.compile(r"((?:[ -<>-~])+)=((?:[ -~])*)")


def autocast_value(value):
    """Cast string to string, float, bool or None. """
    if value is None:
        return
    value_lcase = value.lower()
    if value_lcase == "null":
        return
    if value_lcase == "false":
        return False
    if value_lcase == "true":
        return True
    try:
        return float(value)
    except ValueError:
        return value


class ParseTextColumn(BaseSingle):
    """Parses a text column into (possibly multiple) value columns.

    Key, value pairs need to be separated by an equal (``=``) sign.

    :param source: data source
    :param source_column: existing column in source.
    :param key_mapping: mapping containing pairs {key_name: column_name}:
        key_name: existing key in the text to be parsed.
        column_name: name of the new column created that contains
        the parsed value.

    :type source: GeometryBlock
    :type source_column: string
    :type key_mapping: dict
    """

    def __init__(self, source, source_column, key_mapping):
        if not isinstance(source, GeometryBlock):
            raise TypeError("'{}' object is not allowed.".format(type(source)))
        if not isinstance(source_column, str):
            raise TypeError("'{}' object is not allowed.".format(type(source_column)))
        if source_column not in source.columns:
            raise KeyError("Column '{}' is not available.".format(source_column))
        if not isinstance(key_mapping, dict):
            raise TypeError("'{}' object is not allowed.".format(type(key_mapping)))
        super().__init__(source, source_column, key_mapping)

    @property
    def source(self):
        return self.args[0]

    @property
    def source_column(self):
        return self.args[1]

    @property
    def key_mapping(self):
        return self.args[2]

    @property
    def columns(self):
        return self.source.columns | set(self.key_mapping.values())

    def get_sources_and_requests(self, **request):
        process_kwargs = {
            "source_column": self.source_column,
            "key_mapping": self.key_mapping,
        }
        return [(self.source, request), (process_kwargs, None)]

    @staticmethod
    def process(data, kwargs):
        source_column = kwargs["source_column"]
        key_mapping = kwargs["key_mapping"]

        if "features" not in data or len(data["features"]) == 0:
            return data  # do nothing for non-feature requests

        f = data["features"].copy()
        # Most of the code in this function revolves around the idea that we do
        # not keep large strings multiple times in memory. We use the pandas
        # categorical for that.

        # Hopefully, this is already the case:
        column = f[source_column].astype("category")

        if len(column.cat.categories) == 0:
            # no data to parse: add empty columns and return directly
            f = f.reindex(columns=list(f.columns) + list(key_mapping.values()))
            return {"features": f, "projection": data["projection"]}

        def parser(description):
            pairs = dict(REGEX_KEYVALUE.findall(description))
            return [autocast_value(pairs.get(key)) for key in key_mapping.keys()]

        # Parse each category
        extra_columns = pd.DataFrame(
            [parser(x) for x in column.cat.categories], columns=key_mapping.values()
        )

        # Convert strings to categoricals before aligning
        for name, dtype in zip(extra_columns.columns, extra_columns.dtypes):
            if dtype == object:
                extra_columns[name] = extra_columns[name].astype("category")

        # Align the generated dataframe with the original. Pandas versions
        # later than 0.19 have a pd.align that could be used also.
        try:
            extra_columns_aligned = extra_columns.loc[column.cat.codes]
            extra_columns_aligned.index = f.index
        except KeyError:
            extra_columns_aligned = pd.DataFrame([], columns=key_mapping.values())

        # Assign the extra columns one by one. This preserves categoricals.
        for name in extra_columns_aligned.columns:
            f[name] = extra_columns_aligned[name]

        return {"features": f, "projection": data["projection"]}



"""
Module containing elementwise raster blocks.
"""
from functools import wraps

import numpy as np

from dask_geomodeling.utils import get_dtype_max, get_index, GeoTransform

from .base import RasterBlock, BaseSingle

__all__ = [
    "Add",
    "Subtract",
    "Multiply",
    "Divide",
    "Power",
    "FillNoData",
    "Equal",
    "NotEqual",
    "Greater",
    "GreaterEqual",
    "Less",
    "LessEqual",
    "Invert",
    "And",
    "Or",
    "Xor",
    "IsData",
    "IsNoData",
]


class BaseElementwise(RasterBlock):
    """ Base block for elementwise operations on rasters.

    This block only accepts stores that have aligned frames. The``extent``
    and ``period`` attributes are limited to the intersection of the
    ancestor stores. If the ancestor stores do not overlap (spatially or
    temporally), the elementwise block will be empty.
    """

    def __init__(self, *args):
        super(BaseElementwise, self).__init__(*args)
        # check the timedelta so that an error is raised if incompatible
        if len(self._sources) > 1:
            self.timedelta  # NOQA

    @property
    def _sources(self):
        # go through the arguments and list the blocks
        return [arg for arg in self.args if isinstance(arg, RasterBlock)]

    def get_sources_and_requests(self, **request):
        start = request.get("start", None)
        stop = request.get("stop", None)

        if start is not None and stop is not None:
            # limit request to self.period so that resulting data is aligned
            period = self.period
            if period is not None:
                request["start"] = max(start, period[0])
                request["stop"] = min(stop, period[1])

        process_kwargs = {"dtype": self.dtype.name, "fillvalue": self.fillvalue}
        sources_and_requests = [(source, request) for source in self.args]

        return [(process_kwargs, None)] + sources_and_requests

    @property
    def timedelta(self):
        """ The period between timesteps in case of equidistant time. """
        if len(self._sources) == 1:
            return self._sources[0].timedelta

        timedeltas = [s.timedelta for s in self._sources]
        if any(timedelta is None for timedelta in timedeltas):
            return None  # None precedes

        # multiple timedeltas: assert that they are equal
        if not timedeltas[1:] == timedeltas[:-1]:
            raise ValueError(
                "Time resolutions of input rasters are not "
                "equal ({}).".format(timedeltas)
            )
        else:
            return timedeltas[0]

    @property
    def period(self):
        """ Return period datetime tuple. """
        if len(self._sources) == 1:
            return self._sources[0].period

        periods = [s.period for s in self._sources]
        if any(period is None for period in periods):
            return None  # None precedes

        # multiple periods: return the overlapping period
        start = max([p[0] for p in periods])
        stop = min([p[1] for p in periods])
        if stop < start:
            return None  # no overlap
        else:
            return start, stop

    @property
    def extent(self):
        """ Boundingbox of contents in WGS84 projection. """
        if len(self._sources) == 1:
            return self._sources[0].extent

        extents = [s.extent for s in self._sources]
        if any(extent is None for extent in extents):
            return None  # None precedes

        # multiple extents: return the overlapping box
        x1 = max([e[0] for e in extents])
        y1 = max([e[1] for e in extents])
        x2 = min([e[2] for e in extents])
        y2 = min([e[3] for e in extents])
        if x2 <= x1 or y2 <= y1:
            return None  # no overlap
        else:
            return x1, y1, x2, y2

    @property
    def dtype(self):
        dtype = np.result_type(*self.args)
        if np.issubdtype(dtype, np.integer) or dtype == np.bool:
            # use at least int32
            return np.result_type(dtype, np.int32)
        elif np.issubdtype(dtype, np.floating):
            # use at least float32
            return np.result_type(dtype, np.float32)
        else:
            return dtype

    @property
    def fillvalue(self):
        dtype = self.dtype
        if dtype == np.bool:
            return
        else:
            return get_dtype_max(dtype)

    @property
    def geometry(self):
        """Intersection of geometries in the projection of the first store
        geometry. """
        geometries = [x.geometry for x in self._sources]
        if any(x is None for x in geometries):
            return
        if len(geometries) == 1:
            return geometries[0]
        result = geometries[0]
        sr = result.GetSpatialReference()
        for geometry in geometries[1:]:
            if not geometry.GetSpatialReference().IsSame(sr):
                geometry = geometry.Clone()
                geometry.TransformTo(sr)
            result = result.Intersection(geometry)
        if result.GetArea() == 0.0:
            return
        return result

    @property
    def projection(self):
        """Projection of the data if they match, else None"""
        projection = self._sources[0].projection
        if projection is None:
            return
        for arg in self._sources[1:]:
            if projection != arg.projection:
                return
        return projection

    @property
    def geo_transform(self):
        geo_transform = self._sources[0].geo_transform
        if geo_transform is None:
            return
        geo_transform = GeoTransform(geo_transform)
        for arg in self._sources[1:]:
            other = arg.geo_transform
            if other is None or not geo_transform.aligns_with(other):
                return
        return geo_transform


class BaseMath(BaseElementwise):
    """A block that applies basic math to two store-like objects."""

    def __init__(self, a, b):
        for x in (a, b):
            if not isinstance(x, (RasterBlock, np.ndarray, float, int)):
                raise TypeError("'{}' object is not allowed".format(type(x)))
        super(BaseMath, self).__init__(a, b)


class BaseComparison(BaseMath):
    """A block that applies basic comparisons to two store-like objects."""

    @property
    def dtype(self):
        return np.dtype("bool")


class BaseLogic(BaseElementwise):
    """A block that applies basic logical operations to two store-like objects.
    """

    def __init__(self, a, b):
        for x in (a, b):
            if isinstance(x, (RasterBlock, np.ndarray)):
                dtype = x.dtype
                if dtype != np.dtype("bool"):
                    raise TypeError("inputs must have boolean dtypes")
            elif not isinstance(x, bool):
                raise TypeError("'{}' object is not allowed".format(type(x)))
        super(BaseLogic, self).__init__(a, b)

    @property
    def dtype(self):
        return np.dtype("bool")

    @property
    def fillvalue(self):
        return None


def wrap_math_process_func(func):
    """This enables normal math functions to operate only on the data values.
    Nodata is propagated. In case of comparison operators, nodata becomes
    False, except for NotEqual, where it becomes True.

    'meta' and 'time' fields are propagated from the first source. """

    @wraps(func)  # propagates the name and docstring
    def math_process_func(process_kwargs, *args):
        compute_args = []  # the args for the math operation
        # perform the nodata masking manually, as numpy maskedarrays are slow
        nodata_mask = None

        dtype = process_kwargs["dtype"]
        fillvalue = process_kwargs["fillvalue"]

        for data in args:
            if data is None:
                return None
            if not isinstance(data, dict):
                compute_args.append(data)
            elif "time" in data or "meta" in data:
                # return the time / meta right away. assumes there are no
                # mixed requests and that time is aligned
                return data
            elif "values" in data:
                compute_args.append(data["values"])
                # update the nodata mask
                if data["values"].dtype == np.dtype("bool"):
                    continue  # boolean data does not contain nodata values.
                if "no_data_value" not in data:
                    continue
                _nodata_mask = data["values"] == data["no_data_value"]
                if nodata_mask is None:
                    nodata_mask = _nodata_mask
                else:
                    nodata_mask |= _nodata_mask
            else:
                raise TypeError("Cannot apply math function to value {}".format(data))

        if dtype == np.dtype("bool"):
            no_data_value = None
            if func is np.not_equal:
                fillvalue = True
            else:
                fillvalue = False
            func_kwargs = {}
        else:
            func_kwargs = {"dtype": dtype}
            no_data_value = fillvalue

        with np.errstate(all="ignore"):  # suppresses warnings
            result_values = func(*compute_args, **func_kwargs)

        if nodata_mask is not None:
            result_values[nodata_mask] = fillvalue
        return {"no_data_value": no_data_value, "values": result_values}

    return math_process_func


class Add(BaseMath):
    """
    Add a value.

    :param a: Addition parameter a
    :param b: Addition parameter b

    :type a: RasterBlock, scalar
    :type b: RasterBlock, scalar

    At least one of the parameters should be a RasterBlock. If the
    params are both RasterBlocks, they should share exactly the
    same time structure. The Snap block can be used to accomplish this.
    """

    process = staticmethod(wrap_math_process_func(np.add))


class Subtract(BaseMath):
    """
    Subtract a constant value from a store or vice versa.

    :param a: Subtraction parameter a
    :param b: Subtraction parameter b

    :type a: RasterBlock, scalar
    :type b: RasterBlock, scalar

    At least one of the parameters should be a RasterBlock. If the
    params are both RasterBlocks, they should share exactly the
    same time structure. The Snap block can be used to accomplish this.
    """

    process = staticmethod(wrap_math_process_func(np.subtract))


class Multiply(BaseMath):
    """
    Multiply by a value.

    :param a: Multiplication parameter a
    :param b: Multiplication parameter b

    :type a: RasterBlock, scalar
    :type b: RasterBlock, scalar

    At least one of the parameters should be a RasterBlock. If the
    params are both RasterBlocks, they should share exactly the
    same time structure. The Snap block can be used to accomplish this.
    """

    process = staticmethod(wrap_math_process_func(np.multiply))


class Divide(BaseMath):
    """
    Divide Store by a constant value or vice versa.

    :param a: Division parameter a
    :param b: Division parameter b

    :type a: RasterBlock, scalar
    :type b: RasterBlock, scalar

    At least one of the parameters should be a RasterBlock. If the
    params are both RasterBlocks, they should share exactly the
    same time structure. The Snap block can be used to accomplish this.
    """

    process = staticmethod(wrap_math_process_func(np.divide))

    @property
    def dtype(self):
        # use at least float32
        return np.result_type(np.float32, *self.args)


class Power(BaseMath):
    """
    Raise each number in a to the power b.

    :param a: base
    :param b: exponent

    :type a: RasterBlock, scalar
    :type b: RasterBlock, scalar

    At least one of the parameters should be a RasterBlock. If the
    params are both RasterBlocks, they should share exactly the
    same time structure. The Snap block can be used to accomplish this.
    """

    process = staticmethod(wrap_math_process_func(np.power))

    def __init__(self, a, b):
        # cast negative integer exponents to float, because if a is also of
        # integer type, power will raise a ValueError
        if isinstance(b, int) and b < 0:
            b = float(b)
        super(Power, self).__init__(a, b)


class Equal(BaseComparison):
    """
    Compares the values of two stores and returns True if they are equal.

    Note that "no data" is not equal to "no data".

    :param a: Comparison parameter a
    :param b: Comparison parameter b

    :type a: RasterBlock, scalar
    :type b: RasterBlock, scalar

    At least one of the parameters should be a RasterBlock. If the
    params are both RasterBlocks, they should share exactly the
    same time structure. The Snap block can be used to accomplish this.
    """

    process = staticmethod(wrap_math_process_func(np.equal))


class NotEqual(BaseComparison):
    """
    Compares the values of two stores and returns False if they are equal.

    Note that "no data" is not equal to "no data".

    :param a: Comparison parameter a
    :param b: Comparison parameter b

    :type a: RasterBlock, scalar
    :type b: RasterBlock, scalar

    At least one of the parameters should be a RasterBlock. If the
    params are both RasterBlocks, they should share exactly the
    same time structure. The Snap block can be used to accomplish this.
    """

    process = staticmethod(wrap_math_process_func(np.not_equal))


class Greater(BaseComparison):
    """
    Returns True if a is greater than b.

    Note that "no data" will always return False

    :param a: Comparison parameter a
    :param b: Comparison parameter b

    :type a: RasterBlock, scalar
    :type b: RasterBlock, scalar

    At least one of the parameters should be a RasterBlock. If the
    params are both RasterBlocks, they should share exactly the
    same time structure. The Snap block can be used to accomplish this.
    """

    process = staticmethod(wrap_math_process_func(np.greater))


class GreaterEqual(BaseComparison):
    """
    Returns True if a is greater than or equal to b.

    Note that "no data" will always return False

    :param a: Comparison parameter a
    :param b: Comparison parameter b

    :type a: RasterBlock, scalar
    :type b: RasterBlock, scalar

    At least one of the parameters should be a RasterBlock. If the
    params are both RasterBlocks, they should share exactly the
    same time structure. The Snap block can be used to accomplish this.
    """

    process = staticmethod(wrap_math_process_func(np.greater_equal))


class Less(BaseComparison):
    """
    Returns True if a is less than b.

    Note that "no data" will always return False

    :param a: Comparison parameter a
    :param b: Comparison parameter b

    :type a: RasterBlock, scalar
    :type b: RasterBlock, scalar

    At least one of the parameters should be a RasterBlock. If the
    params are both RasterBlocks, they should share exactly the
    same time structure. The Snap block can be used to accomplish this.
    """

    process = staticmethod(wrap_math_process_func(np.less))


class LessEqual(BaseComparison):
    """
    Returns True if a is less than or equal to b.

    Note that "no data" will always return False

    :param a: Comparison parameter a
    :param b: Comparison parameter b

    :type a: RasterBlock, scalar
    :type b: RasterBlock, scalar

    At least one of the parameters should be a RasterBlock. If the
    params are both RasterBlocks, they should share exactly the
    same time structure. The Snap block can be used to accomplish this.
    """

    process = staticmethod(wrap_math_process_func(np.less_equal))


class Invert(BaseSingle):
    """
    Swaps False and True ("not x" or "~x").

    :param x: raster data to invert
    :type x: RasterBlock
    """

    def __init__(self, x):
        super(Invert, self).__init__(x)
        if x.dtype != np.dtype("bool"):
            raise TypeError("input block must have boolean dtype")

    @staticmethod
    def process(data):
        if "values" in data:
            return {"values": ~data["values"], "no_data_value": None}
        else:
            return data

    @property
    def dtype(self):
        return np.dtype("bool")


class IsData(BaseSingle):
    """Returns True where raster has data.

    :param store:
    :type store: RasterBlock
    """

    def __init__(self, store):
        if store.dtype == np.dtype("bool"):
            raise TypeError("input block must not have boolean dtype")
        super(IsData, self).__init__(store)

    @staticmethod
    def process(data):
        if data is None or "values" not in data:
            return data
        values = data["values"]
        no_data_value = data["no_data_value"]
        return {"values": values != no_data_value, "no_data_value": None}

    @property
    def dtype(self):
        return np.dtype("bool")

    @property
    def fillvalue(self):
        return None


class IsNoData(IsData):
    """Returns True where raster has no data.

    :param store:
    :type store: RasterBlock
    """

    @staticmethod
    def process(data):
        if data is None or "values" not in data:
            return data
        values = data["values"]
        no_data_value = data["no_data_value"]
        return {"values": values == no_data_value, "no_data_value": None}


class And(BaseLogic):
    """
    Returns True where a and b are True.

    :param a: Comparison parameter a
    :param b: Comparison parameter b

    :type a: RasterBlock, boolean
    :type b: RasterBlock, boolean

    All input parameters should have a boolean dtype and at least one of the
    parameters should be a RasterBlock. If both are RasterBlocks, they should
    share exactly the same time structure. The Snap block can be used to
    accomplish this.
    """

    process = staticmethod(wrap_math_process_func(np.logical_and))


class Or(BaseLogic):
    """
    Returns True where a or b are True.

    :param a: Comparison parameter a
    :param b: Comparison parameter b

    :type a: RasterBlock, boolean
    :type b: RasterBlock, boolean

    All input parameters should have a boolean dtype and at least one of the
    parameters should be a RasterBlock. If both are RasterBlocks, they should
    share exactly the same time structure. The Snap block can be used to
    accomplish this.
    """

    process = staticmethod(wrap_math_process_func(np.logical_or))


class Xor(BaseLogic):
    """
    Returns True where either a or b is True (exclusive-or)

    :param a: Comparison parameter a
    :param b: Comparison parameter b

    :type a: RasterBlock, boolean
    :type b: RasterBlock, boolean

    All input parameters should have a boolean dtype and at least one of the
    parameters should be a RasterBlock. If both are RasterBlocks, they should
    share exactly the same time structure. The Snap block can be used to
    accomplish this.
    """

    process = staticmethod(wrap_math_process_func(np.logical_xor))


class FillNoData(BaseElementwise):
    """
    Combines multiple rasters, filling in nodata values.

    :param args: list of raster sources to be combined.
    :type args: list of RasterBlock

    Values at equal timesteps in the contributing rasters are pasted left to
    right. Therefore values from rasters that are more 'to the left' are
    shadowed by values from rasters more 'to the right'. However, 'no data'
    values are transparent and do not shadow underlying data values.
    """

    def __init__(self, *args):
        # TODO Expand this block so that it takes ndarrays and scalars.
        for arg in args:
            if not isinstance(arg, RasterBlock):
                raise TypeError("'{}' object is not allowed".format(type(arg)))
        super(FillNoData, self).__init__(*args)

    @staticmethod
    def process(*args):
        """Combine data, filling in nodata values."""
        data_list = []
        no_data_values = []
        for data in args:
            if data is None:
                continue
            elif "time" in data or "meta" in data:
                # return the time / meta right away. assumes there are no
                # mixed requests and that time is aligned
                return data
            elif "values" in data and "no_data_value" in data:
                data_list.append(data["values"])
                no_data_values.append(data["no_data_value"])

        dtype = np.result_type(*data_list)
        fillvalue = get_dtype_max(dtype)

        # initialize values array
        values = np.full(data_list[0].shape, fillvalue, dtype=dtype)

        # populate values array
        for data, no_data_value in zip(data_list, no_data_values):
            # index is where the source has data
            index = get_index(data, no_data_value)
            values[index] = data[index]

        return {"values": values, "no_data_value": fillvalue}



"""
Module containing raster blocks for spatial operations.
"""
import math

from scipy import ndimage
import numpy as np

from dask_geomodeling.utils import (
    EPSG3857,
    get_sr,
    Extent,
    get_dtype_min,
    get_footprint,
)

from .base import BaseSingle

__all__ = ["Dilate", "Smooth", "MovingMax", "HillShade"]


def expand_request_pixels(request, radius=1):
    """ Expand request by `radius` pixels. Returns None for non-vals requests
    or point requests. """
    if request["mode"] != "vals":  # do nothing with time and meta requests
        return None

    width, height = request["width"], request["height"]
    x1, y1, x2, y2 = request["bbox"]
    pwidth, pheight = x2 - x1, y2 - y1

    if pwidth == 0 or pheight == 0:  # cannot dilate a point request
        return None

    amount_x = pwidth / width * radius
    amount_y = pheight / height * radius

    new_request = request.copy()
    new_request["bbox"] = (x1 - amount_x, y1 - amount_x, x2 + amount_y, y2 + amount_y)
    new_request["width"] += 2 * radius
    new_request["height"] += 2 * radius
    return new_request


def expand_request_meters(request, radius_m=1):
    """
    Expand request by `radius_m` meters, rounded so that an integer number of
    pixels is added to all sides.

    Returns a tuple of:
     - new request with adapted bbox, width and height
     - the radius transformed to pixels as a (y, x) tuple of floats
     - the added margins as a (y, x) tuple of integers
    """
    sr = get_sr(request["projection"])
    bbox = request["bbox"]

    # throughout, variables in the projected unit ( = meters, mostly) are
    # suffixed by _m, in pixels by _px
    if sr.IsGeographic():
        # expand geographic bbox in EPSG3857
        extent_geom = Extent(bbox, sr)
        bbox = extent_geom.transformed(EPSG3857).bbox
    else:
        # most Projected projections are in meters, but to be sure:
        radius_m /= sr.GetLinearUnits()

    # compute the initial zoom factors: how much to expand the bbox to obtain
    # margins of exactly 'radius' (in meters)
    x1, y1, x2, y2 = bbox
    shape_m = y2 - y1, x2 - x1

    # omit the +1 in zoom for efficiency: zoom=0.2 is a zoom factor of 1.2
    zoom = [2 * radius_m / s for s in shape_m]

    # compute the size in pixels, and compute the margins (rounded size)
    shape_px = request["height"], request["width"]
    radius_px = [z * s / 2 for (z, s) in zip(zoom, shape_px)]
    margins_px = [int(round(sz)) for sz in radius_px]

    # use these (int-valued) margins to compute the actual zoom and margins
    zoom = [2 * m / s for (s, m) in zip(shape_px, margins_px)]
    margins_m = [z * s / 2 for (z, s) in zip(zoom, shape_m)]

    # assemble the request
    new_request = request.copy()
    new_request["bbox"] = (
        x1 - margins_m[1],
        y1 - margins_m[0],
        x2 + margins_m[1],
        y2 + margins_m[0],
    )
    if sr.IsGeographic():
        # transform back to original projection
        extent_proj = Extent(new_request["bbox"], EPSG3857)
        new_request["bbox"] = extent_proj.transformed(sr).bbox
    new_request["height"] += 2 * margins_px[0]
    new_request["width"] += 2 * margins_px[1]

    return new_request, radius_px


class Dilate(BaseSingle):
    """
    Perform binary dilation on specific values. Dilation is done in the
    order of the values parameter.

    :param store: raster to perform dilation on
    :param values: only dilate pixels that have these values

    :type store: RasterBlock
    :type values: list
    """

    def __init__(self, store, values):
        values = np.asarray(values, dtype=store.dtype)
        super(Dilate, self).__init__(store, values)

    @property
    def values(self):
        return self.args[1]

    def get_sources_and_requests(self, **request):
        new_request = expand_request_pixels(request, radius=1)
        if new_request is None:  # not an expandable request: do nothing
            return [(self.store, request)]
        else:
            return [(self.store, new_request), (self.values, None)]

    @staticmethod
    def process(data, values=None):
        if data is None or values is None or "values" not in data:
            return data
        original = data["values"]
        dilated = original.copy()
        for value in values:
            dilated[ndimage.binary_dilation(original == value)] = value
        dilated = dilated[:, 1:-1, 1:-1]
        return {"values": dilated, "no_data_value": data["no_data_value"]}


class MovingMax(BaseSingle):
    """
    Apply a spatial maximum filter to the data using a circular footprint.

    :param store: raster to apply the maximum filter on
    :param size: diameter of the circular footprint in pixels

    :type store: RasterBlock
    :type store: int

    This block can be used for visualization of sparse data.
    """

    def __init__(self, store, size):
        # round size to nearest odd integer
        size = int(2 * round((size - 1) / 2) + 1)
        if size < 3:
            raise ValueError("The size should be odd and larger than 1")
        super(MovingMax, self).__init__(store, size)

    @property
    def size(self):
        return self.args[1]

    def get_sources_and_requests(self, **request):
        size = self.size
        new_request = expand_request_pixels(request, radius=int(size // 2))
        if new_request is None:  # not an expandable request: do nothing
            return [(self.store, request)]
        else:
            return [(self.store, new_request), (size, None)]

    @staticmethod
    def process(data, size=None):
        if data is None or size is None or "values" not in data:
            return data
        radius = int(size // 2)
        footprint = get_footprint(size)[np.newaxis]

        # put absolute minimum on no data pixels
        array = data["values"].copy()
        minimum = get_dtype_min(array.dtype)
        no_data_mask = array == data["no_data_value"]
        array[no_data_mask] = minimum

        # apply maximum filter
        filtered = ndimage.maximum_filter(array, footprint=footprint)

        # replace absolute minimum with original fillvalue
        filtered[(filtered == minimum) & no_data_mask] = data["no_data_value"]

        # cut out the result
        filtered = filtered[:, radius:-radius, radius:-radius]
        return {"values": filtered, "no_data_value": data["no_data_value"]}


class Smooth(BaseSingle):
    """
    Smooth the values from a raster spatially using gaussian smoothing.

    :param store: raster to smooth
    :param size: size of the smoothing in meters. the 'sigma' of the
        gaussian kernel equals ``size / 3``.
    :param fill: fill value to be used for 'no data' values during
        smoothing. The output will not have 'no data' values.

    :type store: Store
    :type size: scalar
    :type fill: scalar

    The challenge is to mitigate the edge effects whilst remaining
    performant. If the necessary margin for smoothing is more than 6 pixels in
    any direction, the approach used here is requesting a zoomed out geometry,
    smooth that and zoom back in to the original region of interest.
    """

    MARGIN_THRESHOLD = 6

    def __init__(self, store, size, fill=0):
        for x in (size, fill):
            if not isinstance(x, (int, float)):
                raise TypeError("'{}' object is not allowed".format(type(x)))
        super(Smooth, self).__init__(store, size, fill)

    @property
    def size(self):
        return self.args[1]

    @property
    def fill(self):
        return self.args[2]

    def get_sources_and_requests(self, **request):
        if request["mode"] != "vals":  # do nothing with time and meta requests
            return [(self.store, request)]

        new_request, size = expand_request_meters(request, self.size)

        # check how many pixels will be added by the request
        if any([s > self.MARGIN_THRESHOLD for s in size]):
            smooth_mode = "zoom"
            # rescale the size
            zoom = [new_request[x] / request[x] for x in ("height", "width")]
            size = [s / z for s, z in zip(size, zoom)]
            # request the original (not expanded) shape
            new_request["height"] = request["height"]
            new_request["width"] = request["width"]
        else:
            smooth_mode = "exact"

        process_kwargs = dict(smooth_mode=smooth_mode, fill=self.fill, size=size)

        return [(self.store, new_request), (process_kwargs, None)]

    @staticmethod
    def process(data, process_kwargs=None):
        if data is None or process_kwargs is None:
            return data
        smooth_mode = process_kwargs["smooth_mode"]
        size_px = process_kwargs["size"]
        fill = process_kwargs["fill"]

        # fill in nodata values
        values = data["values"].copy()
        no_data_value = data["no_data_value"]
        values[values == no_data_value] = fill

        # compute the sigma
        sigma = 0, size_px[0] / 3, size_px[1] / 3
        ndimage.gaussian_filter(
            values, sigma, output=values, mode="constant", cval=fill
        )

        # remove the margins
        if smooth_mode == "exact":
            my, mx = [int(round(s)) for s in size_px]
            values = values[:, my : values.shape[1] - my, mx : values.shape[2] - mx]
        else:
            _, ny, nx = values.shape
            zy, zx = [1 - 2 * size_px[0] / ny, 1 - 2 * size_px[1] / nx]

            values = ndimage.affine_transform(
                values,
                order=0,
                matrix=np.diag([1, zy, zx]),
                offset=[0, size_px[0], size_px[1]],
            )

        return {"values": values, "no_data_value": no_data_value}


class HillShade(BaseSingle):
    """
    Calculate a hillshade from the raster values.

    :param store: RasterBlock object
    :param size: size of the effect, in projected units
    :param altitude: Light source altitude in degrees. Default 45.
    :param azimuth: Light source azimuth in degrees. Default 315.
    :param fill: fill value to be used for 'no data' values during hillshading

    :type store: RasterBlock
    :type size: scalar
    :type altitude: scalar
    :type azimuth: scalar
    :type fill: scalar

    Hillshade, adapted from gdal algorithm. Smooths prior to shading to
    keep it good looking even when zoomed beyond 1:1.
    """

    def __init__(self, store, altitude=45, azimuth=315, fill=0):
        for x in (altitude, azimuth, fill):
            if not isinstance(x, (int, float)):
                raise TypeError("'{}' object is not allowed".format(type(x)))
        super(HillShade, self).__init__(store, float(altitude), float(azimuth), fill)

    @property
    def altitude(self):
        return self.args[1]

    @property
    def azimuth(self):
        return self.args[2]

    @property
    def fill(self):
        return self.args[3]

    @property
    def dtype(self):
        return np.dtype("u1")

    @property
    def fillvalue(self):
        return 256  # on purpose, it does not exist in bytes

    @staticmethod
    def process(data, process_kwargs=None):
        """
        Adapted from:
        https://github.com/OSGeo/gdal/blob/2.0/gdal/apps/gdaldem.cpp#L481

        Edges are not implemented, result clips one pixel from array.
        """
        if process_kwargs is None:
            return data

        array = data["values"].copy()
        array[array == data["no_data_value"]] = process_kwargs["fill"]

        xres, yres = process_kwargs["resolution"]
        alt = math.radians(process_kwargs["altitude"])
        az = math.radians(process_kwargs["azimuth"])
        zsf = 1 / 8  # vertical scale factor
        square_zsf = zsf * zsf

        # gradient
        s0 = slice(None, None), slice(None, -2), slice(None, -2)
        s1 = slice(None, None), slice(None, -2), slice(1, -1)
        s2 = slice(None, None), slice(None, -2), slice(2, None)
        s3 = slice(None, None), slice(1, -1), slice(None, -2)
        s4 = slice(None, None), slice(1, -1), slice(1, -1)
        s5 = slice(None, None), slice(1, -1), slice(2, None)
        s6 = slice(None, None), slice(2, None), slice(None, -2)
        s7 = slice(None, None), slice(2, None), slice(1, -1)
        s8 = slice(None, None), slice(2, None), slice(2, None)

        # angle calculation
        y = np.empty(array.shape, dtype="f4")
        y[s4] = (
            array[s0]
            + 2 * array[s1]
            + array[s2]
            - array[s6]
            - 2 * array[s7]
            - array[s8]
        ) / yres

        x = np.empty(array.shape, dtype="f4")
        x[s4] = (
            array[s0]
            + 2 * array[s3]
            + array[s6]
            - array[s2]
            - 2 * array[s5]
            - array[s8]
        ) / xres

        with np.errstate(all="ignore"):
            xx_plus_yy = x * x + y * y
            aspect = np.arctan2(y, x)

            # shading
            cang = (
                math.sin(alt)
                - math.cos(alt) * zsf * np.sqrt(xx_plus_yy) * np.sin(aspect - az)
            ) / np.sqrt(1 + square_zsf * xx_plus_yy)

        cang = cang[..., 1:-1, 1:-1]
        result = np.where(cang <= 0, 0, 255 * cang).astype("u1")
        return {"values": result, "no_data_value": 256}

    def get_sources_and_requests(self, **request):
        new_request = expand_request_pixels(request, radius=1)
        if new_request is None:  # not an expandable request: do nothing
            return [(self.store, request)]

        # determine resolution
        bbox = request["bbox"]
        resolution = (
            (bbox[2] - bbox[0]) / request["width"],
            (bbox[3] - bbox[1]) / request["height"],
        )

        process_kwargs = dict(
            resolution=resolution,
            altitude=self.altitude,
            azimuth=self.azimuth,
            fill=self.fill,
        )

        return [(self.store, new_request), (process_kwargs, None)]



"""
Module containing miscellaneous raster blocks.
"""
import numpy as np

import shapely

from dask import config
from dask_geomodeling.geometry import GeometryBlock
from dask_geomodeling.utils import (
    get_uint_dtype,
    get_dtype_max,
    get_index,
    rasterize_geoseries,
)

from .base import RasterBlock, BaseSingle


__all__ = ["Clip", "Classify", "Reclassify", "Mask", "MaskBelow", "Step", "Rasterize"]


class Clip(BaseSingle):
    """
    Make result values 'no data' if it is 'no data' in the source.
    If source is a boolean mask, False values become 'no data'.

    :param store: The store-like whose values are to be converted.
    :param source: The store-like as the source for 'data or no data'

    :type store: RasterBlock
    :type source: RasterBlock
    """

    def __init__(self, store, source):
        if not isinstance(source, RasterBlock):
            raise TypeError("'{}' object is not allowed".format(type(store)))
        super(Clip, self).__init__(store, source)

    @property
    def source(self):
        return self.args[1]

    @staticmethod
    def process(data, source_data):
        """ Mask store_data where source_data has no data """
        if data is None:
            return None

        if "values" not in data:
            return data

        # check if values contain data
        if np.all(data["values"] == data["no_data_value"]):
            return data

        # make the boolean mask
        if source_data is None:
            return None

        if source_data["values"].dtype == np.dtype("bool"):
            mask = ~source_data["values"]
        else:
            mask = source_data["values"] == source_data["no_data_value"]

        # adjust values
        values = data["values"].copy()
        values[mask] = data["no_data_value"]
        return {"values": values, "no_data_value": data["no_data_value"]}

    @property
    def extent(self):
        """Intersection of bounding boxes of 'store' and 'source'. """
        result, mask = [s.extent for s in self.args]
        if result is None or mask is None:
            return

        # return the overlapping box
        x1 = max(result[0], mask[0])
        y1 = max(result[1], mask[1])
        x2 = min(result[2], mask[2])
        y2 = min(result[3], mask[3])
        if x2 <= x1 or y2 <= y1:
            return None  # no overlap
        else:
            return x1, y1, x2, y2

    @property
    def geometry(self):
        """Intersection of geometries of 'store' and 'source'. """
        result, mask = [x.geometry for x in self.args]
        if result is None or mask is None:
            return
        sr = result.GetSpatialReference()
        if not mask.GetSpatialReference().IsSame(sr):
            mask = mask.Clone()
            mask.TransformTo(sr)
        result = result.Intersection(mask)
        if result.GetArea() == 0.0:
            return
        return result


class Mask(BaseSingle):
    """
    Convert 'data' values to a single constant value.

    :param store: The store whose values are to be converted.
    :param value: The constant value to be given to 'data' values.

    :type store: Store
    :type value: number
    """

    def __init__(self, store, value):
        if not isinstance(value, (float, int)):
            raise TypeError("'{}' object is not allowed".format(type(value)))
        super(Mask, self).__init__(store, value)

    @property
    def value(self):
        return self.args[1]

    @property
    def fillvalue(self):
        return 1 if self.value == 0 else 0

    @property
    def dtype(self):
        return "float32" if isinstance(self.value, float) else "uint8"

    @staticmethod
    def process(data, value):
        if data is None or "values" not in data:
            return data

        index = get_index(values=data["values"], no_data_value=data["no_data_value"])

        fillvalue = 1 if value == 0 else 0
        dtype = "float32" if isinstance(value, float) else "uint8"

        values = np.full_like(data["values"], fillvalue, dtype=dtype)
        values[index] = value
        return {"values": values, "no_data_value": fillvalue}


class MaskBelow(BaseSingle):
    """
    Mask data below some value.

    :param store: The store whose values are to be masked.
    :param value: The threshold value. Values below this will be masked.

    :type store: Store
    :type value: number
    """

    def __init__(self, store, value):
        if not isinstance(value, (float, int)):
            raise TypeError("'{}' object is not allowed".format(type(value)))
        super(MaskBelow, self).__init__(store, value)

    @staticmethod
    def process(data, value):
        if data is None or "values" not in data:
            return data
        values, no_data_value = data["values"].copy(), data["no_data_value"]
        values[values < value] = no_data_value
        return {"values": values, "no_data_value": no_data_value}


class Step(BaseSingle):
    """Block for a constant step function with one discontinuity.

    The step function is defined as:
    - left if x < location
    - at if x == location
    - right if x > location

    :param store: The raster whose values are the input to the step function.
    :param left: value left of the discontinuity
    :param right: value right of the discontinuity
    :param location: location of the discontinuity
    :param at: value at the discontinuity

    :type store: RasterBlock
    :type left: number
    :type right: number
    :type location: number
    :type at: number
    """

    def __init__(self, store, left=0, right=1, location=0, at=None):
        """Constructor.



        The at parameter defaults to the mean of the left and right values.
        """
        at = (left + right) / 2 if at is None else at
        for x in left, right, location, at:
            if not isinstance(x, (float, int)):
                raise TypeError("'{}' object is not allowed".format(type(x)))
        super(Step, self).__init__(store, left, right, location, at)

    @property
    def left(self):
        return self.args[1]

    @property
    def right(self):
        return self.args[2]

    @property
    def location(self):
        return self.args[3]

    @property
    def at(self):
        return self.args[4]

    @staticmethod
    def process(data, left, right, location, at):
        if data is None or "values" not in data:
            return data

        values, no_data_value = data["values"].copy(), data["no_data_value"]

        # determine boolean index arrays
        mask = values == no_data_value
        left_index = values < location
        at_index = values == location
        right_index = values > location
        # perform mapping
        values[left_index] = left
        values[at_index] = at
        values[right_index] = right
        # put no data values back
        values[mask] = no_data_value

        return {"values": values, "no_data_value": no_data_value}


class Classify(BaseSingle):
    """
    Classify raster data into a binned categories

    :param store: rasterdata to classify
    :param bins: a 1-dimensional and monotonic list of bin edges
    :param right: whether the intervals include the right or the left bin edge

    :type store: RasterBlock
    :type bins: list
    :type right: boolean

    See also:
      https://docs.scipy.org/doc/numpy/reference/generated/numpy.digitize.html
    """

    def __init__(self, store, bins, right=False):
        if not isinstance(store, RasterBlock):
            raise TypeError("'{}' object is not allowed".format(type(store)))
        if not hasattr(bins, "__iter__"):
            raise TypeError("'{}' object is not allowed".format(type(bins)))
        bins_arr = np.asarray(bins)
        if bins_arr.ndim != 1:
            raise TypeError("'bins' should be one-dimensional")
        if not np.issubdtype(bins_arr.dtype, np.number):
            raise TypeError("'bins' should be numeric")
        bins_diff = np.diff(bins)
        if not np.all(bins_diff > 0) or np.all(bins_diff < 0):
            raise TypeError("'bins' should be monotonic")
        super(Classify, self).__init__(store, bins_arr.tolist(), right)

    @property
    def bins(self):
        return self.args[1]

    @property
    def right(self):
        return self.args[2]

    @property
    def dtype(self):
        # with 254 bin edges, we have 255 bins, and we need 256 possible values
        # to include no_data
        return get_uint_dtype(len(self.bins) + 2)

    @property
    def fillvalue(self):
        return get_dtype_max(self.dtype)

    @staticmethod
    def process(data, bins, right):
        if data is None or "values" not in data:
            return data

        values = data["values"]
        dtype = get_uint_dtype(len(bins) + 2)
        fillvalue = get_dtype_max(dtype)

        result_values = np.digitize(values, bins, right).astype(dtype)
        result_values[values == data["no_data_value"]] = fillvalue

        return {"values": result_values, "no_data_value": fillvalue}


class Reclassify(BaseSingle):
    """
    Reclassify integer data to integers or floats.

    :param store: rasterdata to reclassify
    :param data: list of (from, to) values defining the reclassification
        the from values can be of bool or int datatype; the to values can be
        of int or float datatype
    :param select: leave only reclassified values, set others to 'no data'.
      Default False.

    :type store: RasterBlock
    :type data: list
    :type select: bool
    """

    def __init__(self, store, data, select=False):
        dtype = store.dtype
        if dtype != np.bool and not np.issubdtype(dtype, np.integer):
            raise TypeError("The store must be of boolean or integer datatype")

        # validate "data"
        if not hasattr(data, "__iter__"):
            raise TypeError("'{}' object is not allowed".format(type(data)))
        try:
            source, target = map(np.asarray, zip(*data))
        except ValueError:
            raise ValueError("Please supply a list of [from, to] values")
        # "from" can have bool or int dtype, "to" can also be float
        if source.dtype != np.bool and not np.issubdtype(source.dtype, np.integer):
            raise TypeError(
                "Cannot reclassify from value with type '{}'".format(source.dtype)
            )
        if len(np.unique(source)) != len(source):
            raise ValueError("There are duplicates in the reclassify values")
        if not np.issubdtype(target.dtype, np.number):
            raise TypeError(
                "Cannot reclassify to value with type '{}'".format(target.dtype)
            )
        # put 'data' into a list with consistent dtypes
        data = [list(x) for x in zip(source.tolist(), target.tolist())]

        if select is not True and select is not False:
            raise TypeError("'{}' object is not allowed".format(type(select)))
        super().__init__(store, data, select)

    @property
    def data(self):
        return self.args[1]

    @property
    def select(self):
        return self.args[2]

    @property
    def dtype(self):
        _, target = map(np.asarray, zip(*self.data))
        return target.dtype

    @property
    def fillvalue(self):
        return get_dtype_max(self.dtype)

    def get_sources_and_requests(self, **request):
        process_kwargs = {
            "dtype": self.dtype.str,
            "fillvalue": self.fillvalue,
            "data": self.data,
            "select": self.select,
        }
        return [(self.store, request), (process_kwargs, None)]

    @staticmethod
    def process(store_data, process_kwargs):
        if store_data is None or "values" not in store_data:
            return store_data

        no_data_value = store_data["no_data_value"]
        values = store_data["values"]
        source, target = map(np.asarray, zip(*process_kwargs["data"]))
        dtype = np.dtype(process_kwargs["dtype"])
        fillvalue = process_kwargs["fillvalue"]

        # add the nodata value to the source array and map it to the target
        # nodata
        if no_data_value is not None and no_data_value not in source:
            source = np.append(source, no_data_value)
            target = np.append(target, fillvalue)

        # sort the source and target values
        inds = np.argsort(source)
        source = source[inds]
        target = target[inds]

        # create the result array
        if process_kwargs["select"]:  # select = True: initialize with nodata
            result = np.full(values.shape, fillvalue, dtype=dtype)
        else:  # select = True: initialize with existing data
            result = values.astype(dtype)  # makes a copy

        # find all values in the source data that are to be mapped
        mask = np.in1d(values.ravel(), source)
        mask.shape = values.shape
        # place the target values (this also maps nodata values)
        result[mask] = target[np.searchsorted(source, values[mask])]
        return {"values": result, "no_data_value": fillvalue}


class Rasterize(RasterBlock):
    """Converts geometry source to raster

    :param source: geometry source
    :param column_name: column from the geometry source to rasterize. If
      column_name is not provided, a boolean raster will be returned indicating
      where there are geometries.
    :param dtype: a numpy datatype specification to return the array. Defaults
      to 'int32' if column_name is not, else it defaults to 'bool'.
    :param limit: the maximum number of geometries. Defaults to the
      geomodeling.goemetry-limit setting.
    :returns: a raster containing values from 'column_name' or True/False.

    :type source: GeometryBlock
    :type column_name: string
    :type dtype: string
    :type limit: int

    To rasterize floating point values, it is necessary to pass dtype='float'.

    See also:
      https://docs.scipy.org/doc/numpy/reference/arrays.dtypes.html

    The global geometry-limit setting can be adapted as follows:
      >>> from dask import config
      >>> config.set({"geomodeling.geometry-limit": 100000})
    """

    def __init__(self, source, column_name=None, dtype=None, limit=None):
        if not isinstance(source, GeometryBlock):
            raise TypeError("'{}' object is not allowed".format(type(source)))
        if column_name is not None and not isinstance(column_name, str):
            raise TypeError("'{}' object is not allowed".format(type(column_name)))
        if dtype is None:  # set default values
            dtype = "bool" if column_name is None else "int32"
        else:  # parse to numpy dtype and back to string
            dtype = str(np.dtype(dtype))
        if limit and not isinstance(limit, int):
            raise TypeError("'{}' object is not allowed".format(type(limit)))
        if limit and limit < 1:
            raise ValueError("Limit should be greater than 1")
        super(Rasterize, self).__init__(source, column_name, dtype, limit)

    @property
    def source(self):
        return self.args[0]

    @property
    def column_name(self):
        return self.args[1]

    @property
    def limit(self):
        return self.args[3]

    @property
    def dtype(self):
        return np.dtype(self.args[2])

    @property
    def fillvalue(self):
        return None if self.dtype == np.bool else get_dtype_max(self.dtype)

    @property
    def period(self):
        return (self.DEFAULT_ORIGIN,) * 2

    @property
    def extent(self):
        return None

    @property
    def timedelta(self):
        return None

    @property
    def geometry(self):
        return None

    @property
    def projection(self):
        return None

    @property
    def geo_transform(self):
        return None

    def get_sources_and_requests(self, **request):
        # first handle the 'time' and 'meta' requests
        mode = request["mode"]
        if mode == "time":
            return [(self.period[-1], None), ({"mode": "time"}, None)]
        elif mode == "meta":
            return [(None, None), ({"mode": "meta"}, None)]
        elif mode != "vals":
            raise ValueError("Unknown mode '{}'".format(mode))

        # build the request to be sent to the geometry source
        x1, y1, x2, y2 = request["bbox"]
        width, height = request["width"], request["height"]

        # be strict about the bbox, it may lead to segfaults else
        if x2 == x1 and y2 == y1:  # point
            min_size = None
        elif x1 < x2 and y1 < y2:
            min_size = min((x2 - x1) / width, (y2 - y1) / height)
        else:
            raise ValueError("Invalid bbox ({})".format(request["bbox"]))

        limit = self.limit
        if self.limit is None:
            limit = config.get("geomodeling.geometry-limit")

        geom_request = {
            "mode": "intersects",
            "geometry": shapely.geometry.box(*request["bbox"]),
            "projection": request["projection"],
            "min_size": min_size,
            "limit": limit,
            "start": request.get("start"),
            "stop": request.get("stop"),
        }
        # keep some variables for use in process()
        process_kwargs = {
            "mode": "vals",
            "column_name": self.column_name,
            "dtype": self.dtype,
            "no_data_value": self.fillvalue,
            "width": width,
            "height": height,
            "bbox": request["bbox"],
        }
        return [(self.source, geom_request), (process_kwargs, None)]

    @staticmethod
    def process(data, process_kwargs):
        # first handle the time and meta requests
        mode = process_kwargs["mode"]
        if mode == "time":
            return {"time": [data]}
        elif mode == "meta":
            return {"meta": [None]}

        column_name = process_kwargs["column_name"]
        height = process_kwargs["height"]
        width = process_kwargs["width"]
        no_data_value = process_kwargs["no_data_value"]
        dtype = process_kwargs["dtype"]
        f = data["features"]

        # get the value column to rasterize
        if column_name is None:
            values = None
        else:
            try:
                values = f[column_name]
            except KeyError:
                if f.index.name == column_name:
                    values = f.index.to_series()
                else:
                    values = False

        if len(f) == 0 or values is False:  # there is no data to rasterize
            values = np.full((1, height, width), no_data_value, dtype=dtype)
            return {"values": values, "no_data_value": no_data_value}

        result = rasterize_geoseries(
            geoseries=f["geometry"] if "geometry" in f else None,
            values=values,
            bbox=process_kwargs["bbox"],
            projection=data["projection"],
            height=height,
            width=width,
        )

        values = result["values"]

        # cast to the expected dtype if necessary
        cast_values = values.astype(process_kwargs["dtype"])

        # replace the nodata value if necessary
        if result["no_data_value"] != no_data_value:
            cast_values[values == result["no_data_value"]] = no_data_value

        return {"values": cast_values, "no_data_value": no_data_value}



"""
Module containing merge operation that act on geometry blocks
"""
import pandas as pd
from shapely.geometry import box

from .base import GeometryBlock

__all__ = ["MergeGeometryBlocks"]


class MergeGeometryBlocks(GeometryBlock):
    """Merge two GeometryBlocks into one by index

    :param left: left geometry data to merge
    :param right: right geometry data to merge
    :param how: type of merge to be performed. One of ``‘left’, ‘right’,
      ‘outer’, ‘inner'``. Default ``‘inner’``.
    :param suffixes: suffix to apply to overlapping column names in the left
      and right side, respectively. Default ``('', '_right')``.

    :type left: GeometryBlock
    :type right: GeometryBlock
    :type how: string
    :type suffixes: tuple

    See also merge:
      https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.merge.html
    """

    allow_how_joins = ("left", "right", "outer", "inner")

    def __init__(self, left, right, how="inner", suffixes=("", "_right")):
        if not isinstance(left, GeometryBlock):
            raise TypeError("'{}' object is not allowed".format(type(left)))
        if not isinstance(right, GeometryBlock):
            raise TypeError("'{}' object is not allowed".format(type(right)))
        if how not in self.allow_how_joins:
            raise KeyError(
                "'{}' is not part of the list of operations: "
                "{}".format(how, self.allow_how_joins)
            )
        if (
            not isinstance(suffixes[0], str)
            or not isinstance(suffixes[1], str)
            or len(suffixes) != 2
        ):
            raise TypeError("'{}' object is not " "allowed".format(type(suffixes)))
        super().__init__(left, right, how, suffixes)

    @property
    def left(self):
        return self.args[0]

    @property
    def right(self):
        return self.args[1]

    @property
    def how(self):
        return self.args[2]

    @property
    def suffixes(self):
        return self.args[3]

    @property
    def columns(self):
        left = self.left.columns
        right = self.right.columns
        result = left ^ right  # column in left or right, not both
        overlap = left & right
        for col in overlap:
            result |= {col + self.suffixes[0], col + self.suffixes[1]}
        return result

    def get_sources_and_requests(self, **request):
        process_kwargs = {
            "how": self.how,
            "suffixes": self.suffixes,
            "mode": request["mode"],
        }
        return [(self.left, request), (self.right, request), (process_kwargs, None)]

    @staticmethod
    def process(left, right, kwargs):
        mode = kwargs["mode"]
        how = kwargs["how"]
        projection = left["projection"]

        if mode == "intersects" or mode == "centroid":
            merged = pd.merge(
                left["features"],
                right["features"],
                how=kwargs.get("how"),
                suffixes=kwargs.get("suffixes"),
                left_index=True,  # we merge by index, left and right.
                right_index=True,
            )
            return {"features": merged, "projection": projection}
        elif mode == "extent":
            if how == "left":
                return left

            elif how == "right":
                return right

            elif how == "inner":
                values = None
                if left["extent"] and right["extent"]:
                    left_shape = box(*left["extent"])
                    right_shape = box(*right["extent"])
                    extent = left_shape.intersection(right_shape)
                    if not extent.is_empty:
                        values = extent.bounds
                return {"extent": values, "projection": projection}

            elif how == "outer":
                values = None
                if left["extent"] and right["extent"]:
                    left_shape = box(*left["extent"])
                    right_shape = box(*right["extent"])
                    values = left_shape.union(right_shape).bounds
                elif left["extent"]:
                    values = left["extent"]
                elif right["extent"]:
                    values = right["extent"]
                return {"extent": values, "projection": projection}



"""
Module containing the base geometry block classes.
"""
import pandas as pd
from dask_geomodeling import Block

__all__ = ["GeometryBlock", "GetSeriesBlock", "SetSeriesBlock"]


class GeometryBlock(Block):
    """ The base block for geometries

    All geometry blocks must be derived from this base class and must implement
    the following attributes:

    - ``columns``: a set of column names to expect in the dataframe

    A geometry request contains the following fields:

    - mode: one of ``{"intersects", "centroid", "extent"}``
    - geometry: limit returned objects to objects that intersect with this
      shapely geometry object
    - projection: projection to return the geometries in as WKT string
    - limit: the maximum number of geometries
    - min_size: geometries with a bbox that is smaller than this on all sides
      are left out
    - start: start date as UTC datetime
    - stop: stop date as UTC datetime
    - filters: dict of `Django <https://www.djangoproject.com/>`_ ORM-like
      filters on properties (e.g. ``id=598``)

    The data response contains the following:

    - if mode was ``'intersects'``: a DataFrame of features with properties
    - if mode was ``'extent'``: the bbox that contains all features

    To be able to perform operations on properties, there is a helper type
    called``SeriesBlock``. This is the block equivalent of a ``pandas.Series``.
    You can get a ``SeriesBlock`` from a ``GeometryBlock``, perform operations
    on it, and set it back into a ``GeometryBlock``.
    """

    def __getitem__(self, name):
        return GetSeriesBlock(self, name)

    def __setitem__(self, *args, **kwargs):
        raise NotImplementedError("Please use block.set to set a column.")

    def set(self, *args):
        # NB cannot use __setitem__ as block instances are immutable
        return SetSeriesBlock(self, *args)

    def to_file(self, *args, **kwargs):
        """Utility function to export data from this block to a file on disk.

        You need to specify the target file path as well as the extent geometry
        you want to save.

        Args:
          url (str): The target file path. The extension determines the format.
            For supported formats, consult
            GeometryFileSink.supported_extensions.
          fields (dict): a mapping that relates column names to output file
            field names field names,
            ``{<output file field name>: <column name>, ...}``.
          tile_size (int): Optionally use this for large exports to stay within
            memory constraints. The export is split in tiles of given size
            (units are determined by the projection). Finally the tiles are
            merged.
          geometry (shapely Geometry): Limit exported objects to objects whose
            centroid intersects with this geometry.
          projection (str): The projection as a WKT string or EPSG code.
            Sets the projection of the geometry argument, the target
            projection of the data, and the tiling projection.
          start (datetime): start date as UTC datetime
          stop (datetime): stop date as UTC datetime
          **request: see GeometryBlock request specification

        Relevant settings can be adapted as follows:
          >>> from dask import config
          >>> config.set({"geomodeling.root": '/my/output/data/path'})
          >>> config.set({"temporary_directory": '/my/alternative/tmp/dir'})
        """
        from dask_geomodeling.geometry.sinks import to_file

        return to_file(self, *args, **kwargs)


class SeriesBlock(Block):
    """ A helper block for GeometryBlocks, representing one single field"""

    def __add__(self, other):
        from . import Add

        return Add(self, other)

    def __mul__(self, other):
        from . import Multiply

        return Multiply(self, other)

    def __neg__(self):
        from . import Multiply

        return Multiply(self, -1)

    def __sub__(self, other):
        from . import Subtract

        return Subtract(self, other)

    def __floordiv__(self, other):
        from . import FloorDivide

        return FloorDivide(self, other)

    def __mod__(self, other):
        from . import Modulo

        return Modulo(self, other)

    def __truediv__(self, other):
        from . import Divide

        return Divide(self, other)

    def __pow__(self, other):
        from . import Power

        return Power(self, other)

    def __eq__(self, other):
        from . import Equal

        return Equal(self, other)

    def __ne__(self, other):
        from . import NotEqual

        return NotEqual(self, other)

    def __gt__(self, other):
        from . import Greater

        return Greater(self, other)

    def __ge__(self, other):
        from . import GreaterEqual

        return GreaterEqual(self, other)

    def __lt__(self, other):
        from . import Less

        return Less(self, other)

    def __le__(self, other):
        from . import LessEqual

        return LessEqual(self, other)

    def __invert__(self):
        from . import Invert

        return Invert(self)

    def __and__(self, other):
        from . import And

        return And(self, other)

    def __or__(self, other):
        from . import Or

        return Or(self, other)

    def __xor__(self, other):
        from . import Xor

        return Xor(self, other)


class GetSeriesBlock(SeriesBlock):
    """Get a column from a GeometryBlock.

    :param source: GeometryBlock
    :param name: name of the column to get
    :returns: SeriesBlock containing the property column

    :type source: GeometryBlock
    :type name: string
    """

    def __init__(self, source, name):
        if not isinstance(source, GeometryBlock):
            raise TypeError("'{}' object is not allowed".format(type(source)))
        if not isinstance(name, str):
            raise TypeError("'{}' object is not allowed".format(type(name)))
        if name not in source.columns:
            raise KeyError("Column '{}' is not available".format(name))
        super().__init__(source, name)

    @property
    def source(self):
        return self.args[0]

    @staticmethod
    def process(data, name):
        if "features" not in data or name not in data["features"].columns:
            return pd.Series([])
        return data["features"][name]


class SetSeriesBlock(GeometryBlock):
    """Set one or multiple columns (SeriesBlocks) in a GeometryBlock.

    :param source: source to add the extra columns to
    :param column: name of the column to be set
    :param value: series or constant value to set
    :param args: string, SeriesBlock, ..., repeated multiple times
    :returns: the source GeometryBlock with additional property columns

    :type source: GeometryBlock
    :type column: string
    :type value: SeriesBlock, scalar

    Example:
      >>> SetSeriesBlock(view, 'column_1', series_1, 'column_2', series_2)
    """

    def __init__(self, source, column, value, *args):
        if not isinstance(source, GeometryBlock):
            raise TypeError("'{}' object is not allowed".format(type(source)))
        args = (column, value) + args
        if len(args) % 2 != 0:
            raise ValueError("The number of arguments should be even")
        for column in args[::2]:
            if not isinstance(column, str):
                raise TypeError("'{}' object is not allowed".format(type(column)))
        super().__init__(source, *args)

    @property
    def source(self):
        return self.args[0]

    @property
    def columns(self):
        return self.source.columns | set(self.args[1::2])

    @staticmethod
    def process(data, *col_val_pairs):
        if "features" not in data or len(data["features"]) == 0:
            return data
        features = data["features"].copy()
        for column, value in zip(col_val_pairs[::2], col_val_pairs[1::2]):
            features[column] = value

        return {"features": features, "projection": data["projection"]}


class BaseSingle(GeometryBlock):
    """Baseclass for all geometry blocks that adjust a geometry source"""

    def __init__(self, source, *args):
        if not isinstance(source, GeometryBlock):
            raise TypeError("'{}' object is not allowed".format(type(source)))
        super().__init__(source, *args)

    @property
    def source(self):
        return self.args[0]

    @property
    def columns(self):
        return self.source.columns  # by default, columns is not changed


class BaseSingleSeries(SeriesBlock):
    """Baseclass for all series blocks that adjust a single series source"""

    def __init__(self, source, *args):
        if not isinstance(source, SeriesBlock):
            raise TypeError("'{}' object is not allowed".format(type(source)))
        super().__init__(source, *args)

    @property
    def source(self):
        return self.args[0]



"""
Module containing geometry block operations that act on non-geometry fields
"""
import numpy as np
import pandas as pd
import operator
from .base import GeometryBlock, SeriesBlock, BaseSingleSeries

__all__ = [
    "Classify",
    "ClassifyFromColumns",
    "Add",
    "Subtract",
    "Multiply",
    "Divide",
    "FloorDivide",
    "Power",
    "Modulo",
    "Equal",
    "NotEqual",
    "Greater",
    "GreaterEqual",
    "Less",
    "LessEqual",
    "And",
    "Or",
    "Xor",
    "Invert",
    "Where",
    "Mask",
    "Round",
]


class Classify(BaseSingleSeries):
    """Classify a continuous-valued property into binned categories

    :param source: source data to classify
    :param bins: a 1-dimensional and monotonic list of bins.
      How values outside of the bins are classified, depends on the length of
      the labels. If len(labels) = len(bins) - 1, then values outside of the
      bins are classified to NaN. If len(labels) = len(bins) + 1, then values
      outside of the bins are classified to the first and last elements of the
      labels list.
    :param labels: the labels for the returned bins
    :param right: whether the intervals include the right or the left bin edge

    :type source: SeriesBlock
    :type bins: list
    :type labels: list
    :type right: boolean

    See also:
      https://pandas.pydata.org/pandas-docs/version/0.23.4/generated/pandas.cut.html
    """

    def __init__(self, source, bins, labels, right=True):
        if not isinstance(bins, list):
            raise TypeError("'{}' object is not allowed".format(type(bins)))
        if not isinstance(labels, list):
            raise TypeError("'{}' object is not allowed".format(type(labels)))
        if not isinstance(right, bool):
            raise TypeError("'{}' object is not allowed".format(type(right)))
        bins_arr = np.asarray(bins)
        if bins_arr.ndim != 1:
            raise TypeError("'bins' must be one-dimensional")
        if (np.diff(bins) < 0).any():
            raise ValueError("'bins' must increase monotonically.")
        if len(labels) not in (len(bins) - 1, len(bins) + 1):
            raise ValueError(
                "Expected {} or {} labels, got {}".format(
                    len(bins) - 1, len(bins) + 1, len(labels)
                )
            )
        super().__init__(source, bins, labels, right)

    @property
    def bins(self):
        return self.args[1]

    @property
    def labels(self):
        return self.args[2]

    @property
    def right(self):
        return self.args[3]

    @staticmethod
    def process(series, bins, labels, right):
        open_bounds = len(labels) == len(bins) + 1
        if open_bounds:
            bins = np.concatenate([[-np.inf], bins, [np.inf]])
        result = pd.cut(series, bins, right, labels)
        labels_dtype = pd.Series(labels).dtype
        if labels_dtype.name != "object":
            result = pd.Series(result, dtype=labels_dtype)
        if open_bounds:
            # patch the result, we actually want to classify np.inf
            if right:
                result[series == -np.inf] = labels[0]
            else:
                result[series == np.inf] = labels[-1]
        return result


class ClassifyFromColumns(SeriesBlock):
    """Classify a continuous-valued property based on bins located in different
    columns.

    :param source: geometry source to classify
    :param value_column: the column name that contains values to classify
    :param bin_columns: column names in which the bins are stored.
      The bins values need to be sorted in increasing order.
    :param labels: specifies the labels for the returned bins
    :param right: whether the intervals include the right or the left bin edge
      Default True.

    :type source: GeometryBlock
    :type value_column: string
    :type bin_columns: list
    :type labels: list
    :type right: boolean

    See also:
      :class:`dask_geomodeling.geometry.field_operations.Classify`
    """

    def __init__(self, source, value_column, bin_columns, labels, right=True):
        if not isinstance(source, GeometryBlock):
            raise TypeError("'{}' object is not allowed".format(type(source)))
        if not isinstance(value_column, str):
            raise TypeError("'{}' object is not allowed".format(type(value_column)))
        if not isinstance(bin_columns, list):
            raise TypeError("'{}' object is not allowed".format(type(bin_columns)))
        if not isinstance(labels, list):
            raise TypeError("'{}' object is not allowed".format(type(labels)))
        if not isinstance(right, bool):
            raise TypeError("'{}' object is not allowed".format(type(right)))
        missing_columns = (set(bin_columns) | {value_column}) - source.columns
        if missing_columns:
            raise KeyError("Columns '{}' are not present".format(missing_columns))
        if len(labels) not in (len(bin_columns) - 1, len(bin_columns) + 1):
            raise ValueError(
                "Expected {} or {} labels, got {}".format(
                    len(bin_columns) - 1, len(bin_columns) + 1, len(labels)
                )
            )
        super().__init__(source, value_column, bin_columns, labels, right)

    @property
    def source(self):
        return self.args[0]

    @property
    def value_column(self):
        return self.args[1]

    @property
    def bin_columns(self):
        return self.args[2]

    @property
    def labels(self):
        return self.args[3]

    @property
    def right(self):
        return self.args[4]

    @staticmethod
    def process(data, value_column, bin_columns, labels, right):
        if "features" not in data or len(data["features"]) == 0:
            return pd.Series([])
        features = data["features"]
        values = features[value_column].values
        bins = features[bin_columns].values
        n_bins = len(bin_columns)

        # Check in which bin every value is. because bins may be different for
        # each value, searchsorted is not an option. We assume that bins are
        # sorted in increasing order. Checking that would be costly.
        if right:
            indices = np.sum(values[:, np.newaxis] > bins, axis=1)
        else:
            indices = np.sum(values[:, np.newaxis] >= bins, axis=1)

        # If we have e.g. 2 labels and 3 bins, the outside intervals are closed
        # any index that is 0 or 3 should become -1 (unclassified).
        if len(labels) == n_bins + 1:  # open bounds
            indices[np.isnan(values)] = -1  # else NaN gets classified
        else:  # closed bounds
            indices[indices == n_bins] = 0
            indices -= 1

        # The output of pd.cut is a categorical Series.
        labeled_data = pd.Categorical.from_codes(indices, labels, ordered=True)
        return pd.Series(labeled_data, index=features.index)


class BaseFieldOperation(BaseSingleSeries):
    """Base block for basic operations between series from a GeometryBlock"""

    def __init__(self, source, other):
        if not isinstance(other, (SeriesBlock, int, float, bool)):
            raise TypeError("'{}' object is not allowed".format(type(other)))
        super().__init__(source, other)

    @property
    def other(self):
        return self.args[1]

    @staticmethod
    def process(source, other):
        raise NotImplementedError()  # implement this in subclasses


class Add(BaseFieldOperation):
    """
    Addition of series and other, element-wise.

    :type source: SeriesBlock
    :type other: SeriesBlock, float

    See also:
      https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.Series.add.html
    """

    process = staticmethod(operator.add)


class Subtract(BaseFieldOperation):
    """
    Subtraction of series and other, element-wise.

    :type source: SeriesBlock
    :type other: SeriesBlock, float

    See also:
      https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.Series.subtract.html
    """

    process = staticmethod(operator.sub)


class Multiply(BaseFieldOperation):
    """
    Multiplication of series and other, element-wise.

    :type source: SeriesBlock
    :type other: SeriesBlock, float

    See also:
      https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.Series.multiply.html
    """

    process = staticmethod(operator.mul)


class Divide(BaseFieldOperation):
    """
    Floating division of series and other, element-wise.

    :type source: SeriesBlock
    :type other: SeriesBlock, float

    Putting source in the divisor is not possible: please use the Power
    for that instead.

    See also:
      https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.Series.divide.html
    """

    process = staticmethod(operator.truediv)


class FloorDivide(BaseFieldOperation):
    """
    Integer (floor) division of series and other, element-wise.

    :type source: SeriesBlock
    :type other: SeriesBlock, float

    See also:
      https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.Series.floordiv.html
    """

    process = staticmethod(operator.floordiv)


class Power(BaseFieldOperation):
    """
    Power (exponent) of series and other, element-wise.

    :type source: SeriesBlock
    :type other: SeriesBlock, float

    See also:
      https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.Series.pow.html
    """

    def __init__(self, source, other):
        # the float(other) will raise a TypeError if necessary
        super().__init__(source, float(other))

    process = staticmethod(operator.pow)


class Modulo(BaseFieldOperation):
    """
    Modulo of series and other, element-wise.

    :type source: SeriesBlock
    :type other: SeriesBlock, float

    See also:
      https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.Series.mod.html
    """

    process = staticmethod(operator.mod)


class Equal(BaseFieldOperation):
    """
    Equal to of series and other, element-wise.

    :type source: SeriesBlock
    :type other: SeriesBlock, float

    See also:
      https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.Series.eq.html
    """

    process = staticmethod(operator.eq)


class NotEqual(BaseFieldOperation):
    """
    Not equal to of series and other, element-wise.

    :type source: SeriesBlock
    :type other: SeriesBlock, float

    See also:
      https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.Series.ne.html
    """

    process = staticmethod(operator.ne)


class Greater(BaseFieldOperation):
    """
    Greater than of series and other, element-wise.

    :type source: SeriesBlock
    :type other: SeriesBlock, float

    See also:
      https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.Series.gt.html
    """

    process = staticmethod(operator.gt)


class GreaterEqual(BaseFieldOperation):
    """
    Greater than or equal to of series and other, element-wise.

    :type source: SeriesBlock
    :type other: SeriesBlock, float

    See also:
      https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.Series.ge.html
    """

    process = staticmethod(operator.ge)


class Less(BaseFieldOperation):
    """
    Less than of series and other, element-wise.

    :type source: SeriesBlock
    :type other: SeriesBlock, float

    See also:
      https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.Series.lt.html
    """

    process = staticmethod(operator.lt)


class LessEqual(BaseFieldOperation):
    """
    Less than or equal to of series and other, element-wise.

    :type source: SeriesBlock
    :type other: SeriesBlock, float

    See also:
      https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.Series.le.html
    """

    process = staticmethod(operator.le)


class BaseLogicOperation(BaseFieldOperation):
    """Base block for logic operations between columns in a GeometryBlock"""

    def __init__(self, source, other):
        if not isinstance(other, SeriesBlock):
            raise TypeError("'{}' object is not allowed".format(type(other)))
        super().__init__(source, other)


class And(BaseLogicOperation):
    """
    Logical AND between series and other.

    :type source: SeriesBlock
    :type other: SeriesBlock, float
    """

    process = staticmethod(operator.and_)


class Or(BaseLogicOperation):
    """
    Logical OR between series and other.

    :type source: SeriesBlock
    :type other: SeriesBlock, float
    """

    process = staticmethod(operator.or_)


class Xor(BaseLogicOperation):
    """
    Logical XOR between series and other.

    :type source: SeriesBlock
    :type other: SeriesBlock, float
    """

    process = staticmethod(operator.xor)


class Invert(BaseSingleSeries):
    """
    Logical NOT operation on a series.

    :type source: SeriesBlock
    """

    process = staticmethod(operator.inv)


class Where(BaseSingleSeries):
    """Replace values where the condition is False.

    :param source: source data
    :param cond: condition that determines whether to keep values from source
    :param other: entries where cond is False are replaced with the
      corresponding value from other.

    :type source: SeriesBlock
    :type cond: SeriesBlock
    :type other: SeriesBlock, scalar

    See also:
      https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.Series.where.html
    """

    def __init__(self, source, cond, other):
        if not isinstance(cond, SeriesBlock):
            raise TypeError("'{}' object is not allowed".format(type(cond)))
        super().__init__(source, cond, other)

    @property
    def cond(self):
        return self.args[1]

    @property
    def other(self):
        return self.args[2]

    @staticmethod
    def process(source, cond, other):
        return source.where(cond, other)


class Mask(BaseSingleSeries):
    """Replace values where the condition is True.

    :param source: source data
    :param cond: condition that determines whether to mask values from source
    :param other: entries where cond is True are replaced with the
      corresponding value from other.

    :type source: SeriesBlock
    :type cond: SeriesBlock
    :type other: SeriesBlock, scalar

    See also:
      https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.Series.mask.html
    """

    def __init__(self, source, cond, other):
        if not isinstance(cond, SeriesBlock):
            raise TypeError("'{}' object is not allowed".format(type(cond)))
        super().__init__(source, cond, other)

    @property
    def cond(self):
        return self.args[1]

    @property
    def other(self):
        return self.args[2]

    @staticmethod
    def process(source, cond, other):
        return source.mask(cond, other)


class Round(BaseSingleSeries):
    """Round each value in a SeriesBlock to the given number of decimals

    :param source: source data
    :param decimals: number of decimal places to round to (default: 0).
      If decimals is negative, it specifies the number of positions to the left
      of the decimal point.

    :type source: SeriesBlock
    :type decimals: int

    See also:
      https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.Series.round.html
    """

    def __init__(self, source, decimals=0):
        if not isinstance(decimals, int):
            raise TypeError("'{}' object is not allowed".format(type(decimals)))
        super().__init__(source, decimals)

    process = staticmethod(np.around)


class Interp(BaseSingleSeries):
    """One-dimensional linear interpolation.

    Compute the one-dimensional piecewise linear interpolant to a function with
    given discrete data points (xp, fp).

    :param source: source data (x-coordinates) to interpolate
    :param xp: the x-coordinates of the data points, must be increasing.
    :param fp: the y-coordinates of the data points, same length as ``xp``.
    :param left: value to return for ``x < xp[0]``, default is ``fp[0]``.
    :param right: value to return for ``x > xp[-1]``, default is ``fp[-1]``.

    :type source: SeriesBlock
    :type xp: list
    :type fp: list
    :type left: float
    :type right: float

    See also:
      https://docs.scipy.org/doc/numpy/reference/generated/numpy.interp.html
    """

    def __init__(self, source, xp, fp, left=None, right=None):
        xp = [float(x) for x in xp]
        fp = [float(x) for x in fp]
        if left is not None:
            left = float(left)
        if right is not None:
            right = float(right)
        if np.any(np.diff(xp) < 0):
            raise ValueError("xp must be monotonically increasing")
        super().__init__(source, xp, fp, left, right)

    @staticmethod
    def process(data, xp, fp, left, right):
        result = np.interp(data, xp, fp, left, right)
        return pd.Series(result, index=data.index)


class Choose(BaseSingleSeries):
    """Construct a SeriesBlock from an index series and a multiple series to
    choose from.

    :param source: series having integers from 0 to n - 1 with n being the
      number of choices. Values outside this range will result in NaN.
    :param choices: multiple series containing the values to choose

    :type source: SeriesBlock
    :type choices: list of SeriesBlock

    See also:
      https://docs.scipy.org/doc/numpy/reference/generated/numpy.choose.html
    """

    def __init__(self, source, *choices):
        if not len(choices) >= 2:
            raise ValueError("The number of choices must be greater than one.")
        if not all([isinstance(choice, SeriesBlock) for choice in choices]):
            raise TypeError("All choices must be SeriesBlock objects")
        super().__init__(source, *choices)

    @property
    def choices(self):
        return self.args[1:]

    @staticmethod
    def process(source, *choices):
        result = pd.Series(np.nan, index=source.index)
        for i, choice in enumerate(choices):
            mask = source == i
            result[mask] = choice[source.index[mask]]
        return result



"""
Module containing operations that return series from geometry fields
"""

import pandas as pd

from dask_geomodeling.utils import shapely_transform

from .base import SeriesBlock, GeometryBlock


__all__ = ["Area"]


class Area(SeriesBlock):
    """Block that calculates the area of geometries.

    :param source: geometry data
    :param projection: projection as EPSG or WKT string to compute area in
    :returns: SeriesBlock with only the computed area

    :type source: GeometryBlock
    :type projection: string
    """

    def __init__(self, source, projection):
        if not isinstance(source, GeometryBlock):
            raise TypeError("'{}' object is not allowed".format(type(source)))
        if not isinstance(projection, str):
            raise TypeError("Argument 'projection' must be a str.")
        super().__init__(source, projection)

    @property
    def source(self):
        return self.args[0]

    @property
    def projection(self):
        return self.args[1]

    @staticmethod
    def process(data, projection):
        if "features" not in data or len(data["features"]) == 0:
            return pd.Series([], dtype=float)

        return (
            data["features"]
            .geometry.apply(shapely_transform, args=(data["projection"], projection))
            .area
        )



"""
Module containing geometry block constructive operations
"""
import numbers

from dask_geomodeling.utils import Extent, shapely_transform, transform_extent

from .base import BaseSingle

__all__ = ["Buffer", "Simplify"]


class Buffer(BaseSingle):
    """Buffer geometries.

    :param source: the geometry source
    :param distance: a distance measure in the given projection.
    :param projection: an EPSG or WKT string, e.g. EPSG:28992.
    :param resolution: quarter circle segments. Default is 16.

    :type source: GeometryBlock
    :type distance: float
    :type projection: string
    :type resolution: int
    """

    def __init__(self, source, distance, projection, resolution=16):
        if not isinstance(distance, numbers.Real):
            raise TypeError("Argument 'distance' must be a float or int.")
        if not isinstance(projection, str):
            raise TypeError("Argument 'projection' must be a str.")
        if not isinstance(resolution, int):
            raise TypeError("Argument 'resolution' must be an int.")
        super().__init__(source, distance, projection, resolution)

    @property
    def distance(self):
        """Buffer distance.

        The unit (e.g. m, °) is determined by the projection.
        """
        return self.args[1]

    @property
    def projection(self):
        """Projection used for buffering."""
        return self.args[2]

    @property
    def resolution(self):
        """Buffer resolution."""
        return self.args[3]

    def get_sources_and_requests(self, **request):
        process_kwargs = {
            "distance": self.distance,
            "buf_srs": self.projection,
            "resolution": self.resolution,
        }
        return [(self.source, request), (process_kwargs, None)]

    @staticmethod
    def process(data, kwargs):
        if "features" in data:
            if len(data["features"]) == 0:
                return data
            req_srs = data["projection"]
            buf_srs = kwargs["buf_srs"]
            features = data["features"].set_geometry(
                data["features"]
                .geometry.apply(shapely_transform, args=(req_srs, buf_srs))
                .buffer(distance=kwargs["distance"], resolution=kwargs["resolution"])
                .apply(shapely_transform, args=(buf_srs, req_srs))
            )
            return {"features": features, "projection": req_srs}
        elif "extent" in data:
            if not data["extent"]:
                return data
            req_srs = data["projection"]
            buf_srs = kwargs["buf_srs"]
            distance = kwargs["distance"]
            extent = transform_extent(data["extent"], req_srs, buf_srs)
            extent = Extent(extent, buf_srs).buffered(distance).bbox
            extent = transform_extent(extent, buf_srs, req_srs)
            return {"extent": extent, "projection": req_srs}
        else:
            raise NotImplementedError("Dunno this mode!")


class Simplify(BaseSingle):
    """Simplify geometries up to given tolerance.

    :param source: the geometry source
    :param tolerance: the simplification tolerance. if no tolerance is given,
       the ``min_size`` request param is used.
    :param preserve_topology: whether to preserve topology. Default True.

    :type source: GeometryBlock
    :type tolerance: float
    :type preserve_topology: boolean
    """

    def __init__(self, source, tolerance=None, preserve_topology=True):
        if tolerance is not None:
            tolerance = float(tolerance)
        super().__init__(source, tolerance, bool(preserve_topology))

    @property
    def tolerance(self):
        return self.args[1]

    @property
    def preserve_topology(self):
        return self.args[2]

    def get_sources_and_requests(self, **request):
        process_kwargs = {
            "tolerance": self.tolerance or request.get("min_size") or 0.0,
            "preserve_topology": self.preserve_topology,
        }
        return [(self.source, request), (process_kwargs, None)]

    @staticmethod
    def process(data, kwargs):
        if "features" not in data:
            # basically assumes that the extent will not change during simplify
            return data
        features = data["features"].set_geometry(
            data["features"].geometry.simplify(**kwargs)
        )  # for 'intersects' and 'centroid' mode.
        return {"features": features, "projection": data["projection"]}



"""
Module containing geometry sources.
"""
import fiona
import geopandas as gpd

from dask import config
from dask_geomodeling import utils
from .base import GeometryBlock

# this import is a copy from geopandas.io.files

__all__ = ["GeometryFileSource"]


class GeometryFileSource(GeometryBlock):
    """A geometry source that opens a geometry file from disk.

    :param url: URL to the file. File paths have to be contained inside the
      current root setting. Relative paths are interpreted relative to this
      setting  but internally stored as absolute paths).
    :param layer: the layer_name in the json to use as source. If None,
      the first layer is used.
    :param id_field: the field name to use as unique ID. Default ``'id'``.

    :type path: string
    :type layer: string
    :type id_field: string

    The input of these blocks is by default limited to 10000 geometries.

    Relevant settings can be adapted as follows:
      >>> from dask import config
      >>> config.set({"geomodeling.root": '/my/data/path'})
      >>> config.set({"geomodeling.geometry-limit": 100000})
    """

    def __init__(self, url, layer=None, id_field="id"):
        safe_url = utils.safe_file_url(url)
        super().__init__(safe_url, layer, id_field)

    @property
    def url(self):
        return self.args[0]

    @property
    def layer(self):
        return self.args[1]

    @property
    def id_field(self):
        return self.args[2]

    @property
    def path(self):
        return utils.safe_abspath(self.url)

    @property
    def columns(self):
        # this is exactly how geopandas determines the columns
        with utils.fiona_env(), fiona.open(self.path) as reader:
            properties = reader.meta["schema"]["properties"]
            return set(properties.keys()) | {"geometry"}

    def get_sources_and_requests(self, **request):
        # check the filters: this block does not support lookups
        if request.get("filters") is None:
            request["filters"] = dict()
        if request["filters"]:
            for field, value in request["filters"].items():
                if "__" in field:
                    raise ValueError("Filter '{}' is not supported".format(field))
        mode = request.get("mode", "intersects").lower()
        if mode not in ("extent", "intersects", "centroid"):
            raise ValueError("Unknown mode '{}'".format(mode))
        request["mode"] = mode
        # just pass on the args and request here
        request["layer"] = self.layer
        request["id_field"] = self.id_field
        return [(self.url, None), (request, None)]

    @staticmethod
    def process(url, request):
        path = utils.safe_abspath(url)

        # convert the requested projection to a fiona CRS
        crs = utils.get_crs(request["projection"])

        # convert the requested shapely geometry object to a GeoSeries
        filt_geom = gpd.GeoSeries([request["geometry"]], crs=crs)

        # acquire the data, filtering on the filt_geom bbox
        f = gpd.GeoDataFrame.from_file(path, bbox=filt_geom, layer=request["layer"])
        if len(f) == 0:
            # return directly if there is no data
            if request.get("mode") == "extent":
                return {"projection": request["projection"], "extent": None}
            else:  # this takes modes 'centroid' and 'intersects'
                return {
                    "projection": request["projection"],
                    "features": gpd.GeoDataFrame([]),
                }

        f.set_index(request["id_field"], inplace=True)

        # apply the non-geometry field filters first
        mask = None
        for field, value in request["filters"].items():
            if field not in f.columns:
                continue
            _mask = f[field] == value
            if mask is None:
                mask = _mask
            else:
                mask &= _mask
        if mask is not None:
            f = f[mask]

        # convert the data to the requested crs
        utils.geodataframe_transform(
            f, utils.crs_to_srs(f.crs), request["projection"]
        )

        # compute the bounds of each geometry and filter on min_size
        min_size = request.get("min_size")
        if min_size:
            bounds = f["geometry"].bounds
            widths = bounds["maxx"] - bounds["minx"]
            heights = bounds["maxy"] - bounds["miny"]
            f = f[(widths > min_size) | (heights > min_size)]

        # only return geometries that truly intersect the requested geometry
        if request["mode"] == "centroid":
            f = f[f["geometry"].centroid.within(filt_geom.iloc[0])]
        else:
            f = f[f["geometry"].intersects(filt_geom.iloc[0])]

        if request.get("mode") == "extent":
            return {
                "projection": request["projection"],
                "extent": tuple(f.total_bounds),
            }
        else:  # this takes modes 'centroid' and 'intersects'
            # truncate the number of geometries if necessary
            if request.get("limit") and len(f) > request["limit"]:
                f = f.iloc[: request["limit"]]
            elif request.get("limit") is None:
                global_limit = config.get("geomodeling.geometry-limit")
                if len(f) > global_limit:
                    raise RuntimeError(
                        "The amount of returned geometries exceeded "
                        "the maximum of {} geometries.".format(global_limit)
                    )

            return {"projection": request["projection"], "features": f}
